"""
In this example we solve a scalar *unfitted* PDE problem. As a
discretisation method we use a level set based geometry description and
a Cut (or Fictitious) Finite element method with a Nitsche formulation
to impose boundary conditions. For stability we add a ghost penalty
stabilization.

Domain:
-------
The domain is [-1,1]^2 while the interface is a ringe described by a
level set function. In the discretisation the level set function is
approximated with a piecewise linear interpolation. This approximate
geometry is then mapped by applying a mesh deformation resulting in
a higher order geometry approximation.

PDE problem:
------------
    - (u_xx + u_yy) = f in Omega (where levelset is negative)
                  u = 0 on dOmega (where levelset is zero)

The r.h.s. term f is chosen according to a manufactured solution.

Discretisation:
---------------
* Background finite element space restricted to active domain (CutFEM)

* Nitsche formulation to impose boundary conditions, see. e.g. [1]

* Ghost penalty stabilization to deal with bad cuts (version as in [2])

Implementational aspects:
-------------------------
* Geometry approximation using isoparametric unfitted FEM

* A (sparse) direct solver is applied to solve the arising linear systems.

References:
-----------
All concepts that are used here are explained in the jupyter tutorials
`basics.ipynb`, `intlset.ipynb` and `cutfem.ipynb`.

Literature:
-----------
[1] E. Burman, P. Hansbo, Fictitious domain finite element methods using
    cut elements: II. A stabilized Nitsche method, Appl. Num. Math.
    62(4):328-341, 2012.
[2] J. Preuß, Higher order unfitted isoparametric space-time FEM on
    moving domains. Master's thesis, NAM, University of Göttingen, 2018.
"""

# ------------------------------ LOAD LIBRARIES -------------------------------
from netgen.geom2d import SplineGeometry
from ngsolve import *
from ngsolve.internal import *
from xfem import *
from xfem.lsetcurv import *

ngsglobals.msg_level = 2

# -------------------------------- PARAMETERS ---------------------------------
# Quadrilateral (or simplicial mesh)
quad_mesh = False
# Mesh diameter
maxh = 0.1
# Finite element space order
order = 3
# Stabilization parameter for ghost-penalty
gamma_stab = 0.1
# Stabilization parameter for Nitsche
lambda_nitsche = 10 * order * order


# ----------------------------------- MAIN ------------------------------------
# Geometry and Mesh
square = SplineGeometry()
square.AddRectangle((-1, -1), (1, 1), bc=1)
ngmesh = square.GenerateMesh(maxh=maxh, quad_dominated=quad_mesh)
mesh = Mesh(ngmesh)


# Manufactured exact solution for monitoring the error
r2 = 3 / 4  # outer radius
r1 = 1 / 4  # inner radius
rc = (r1 + r2) / 2.0
rr = (r2 - r1) / 2.0
r = sqrt(x**2 + y**2)
levelset = IfPos(r - rc, r - rc - rr, rc - r - rr)

exact = (20 * (r2 - sqrt(x**2 + y**2)) * (sqrt(x**2 + y**2) - r1)).Compile()
coeff_f = - (exact.Diff(x).Diff(x) + exact.Diff(y).Diff(y)).Compile()

# Higher order level set approximation
lsetmeshadap = LevelSetMeshAdaptation(mesh, order=order, threshold=0.1,
                                      discontinuous_qn=True)
deformation = lsetmeshadap.CalcDeformation(levelset)
lsetp1 = lsetmeshadap.lset_p1


# Element, facet and dof marking w.r.t. boundary approximation with lsetp1:
ci = CutInfo(mesh, lsetp1)
hasneg = ci.GetElementsOfType(HASNEG)
hasif = ci.GetElementsOfType(IF)

# facets used for stabilization:
ba_facets = GetFacetsWithNeighborTypes(mesh, a=hasneg, b=hasif)

Vhbase = H1(mesh, order=order, dirichlet=[], dgjumps=True)
Vh = Restrict(Vhbase, hasneg)

gfu = GridFunction(Vh)

u, v = Vh.TrialFunction(), Vh.TestFunction()
h = specialcf.mesh_size
n = Normalize(grad(lsetp1))

# integration domains:
dx = dCut(lsetp1, NEG, definedonelements=hasneg, deformation=deformation)
ds = dCut(lsetp1, IF, definedonelements=hasif, deformation=deformation)
dw = dFacetPatch(definedonelements=ba_facets, deformation=deformation)

a = BilinearForm(Vh, symmetric=False)
# Diffusion term
a += grad(u) * grad(v) * dx
# Nitsche term
a += -grad(u) * n * v * ds
a += -grad(v) * n * u * ds
a += (lambda_nitsche / h) * u * v * ds
# Ghost penalty stabilization (near the boundary)
a += gamma_stab / h**2 * (u - u.Other()) * (v - v.Other()) * dw

# R.h.s. term:
f = LinearForm(Vh)
f += coeff_f * v * dx

# Assemble system
a.Assemble()
f.Assemble()

# Solve linear system
gfu.vec.data = a.mat.Inverse(Vh.FreeDofs()) * f.vec

# Measure the error
l2error = sqrt(Integrate((gfu - exact)**2 * dx, mesh))
print("L2 Error: {0}".format(l2error))

# visualization:
Draw(deformation, mesh, "deformation")
DrawDC(lsetp1, gfu, 0, mesh, "uh", deformation=deformation)



"""
Basic tests for multiple level-set functionality
"""

# ------------------------------ LOAD LIBRARIES -------------------------------
import pytest
from netgen.geom2d import SplineGeometry
from ngsolve import *
from xfem import *
from xfem.mlset import *

ngsglobals.msg_level = 1


# ----------------------------- UTILITY FUNTIONS ------------------------------
def UpdateMarkers(arr, arr_union, arr_intersection=None):
    arr[:] = False
    arr |= arr_union
    if arr_intersection:
        arr &= arr_intersection


def MarkersEqual(arr1, arr2):
    if len(arr1) != len(arr2):
        return False
    for i in range(len(arr1)):
        if arr1[i] != arr2[i]:
            return False
    return True


# -----------------------------------------------------------------------------
# --------------------------------- 2D TESTS ----------------------------------
# -----------------------------------------------------------------------------


def test_2d_mlci_and_lo_integration():
    # ---------------------------- Background Mesh ----------------------------
    square = SplineGeometry()
    square.AddRectangle([-1, -0.5], [1, 1.5], bc=1)
    mesh = Mesh(square.GenerateMesh(maxh=0.5, quad_dominated=False))

    # ------------------------------ Level Sets -------------------------------
    level_sets = [y - 1, 2 * x - y, -2 * x - y]
    nr_ls = len(level_sets)

    level_sets_p1 = tuple(GridFunction(H1(mesh, order=1)) for i in range(nr_ls))
    for i, lset_p1 in enumerate(level_sets_p1):
        InterpolateToP1(level_sets[i], lset_p1)

    lset_mult = CoefficientFunction(1)
    for lset_p1 in level_sets_p1:
        lset_mult *= IfPos(lset_p1, 0, 1)
    Draw(lset_mult, mesh, "lset_mult")

    # ---------------------------- DomainTypeArray ----------------------------
    triangle = DomainTypeArray((NEG, NEG, NEG))
    Draw(triangle.Indicator(level_sets_p1), mesh, "dta_indicator")

    # ------------------------------- Cut-Info --------------------------------
    mlci = MultiLevelsetCutInfo(mesh, level_sets_p1)

    # ------------------------- Element Marker Tests --------------------------
    els_neg, els_hasneg = BitArray(mesh.ne), BitArray(mesh.ne)
    els_neg2, els_hasneg2 = BitArray(mesh.ne), BitArray(mesh.ne)
    els_not_neg, els_if = BitArray(mesh.ne), BitArray(mesh.ne)

    UpdateMarkers(els_neg, mlci.GetElementsOfType(triangle))
    UpdateMarkers(els_neg2, mlci.GetElementsOfType(triangle.as_list))

    UpdateMarkers(els_hasneg, mlci.GetElementsWithContribution(triangle))
    UpdateMarkers(els_hasneg2,
                  mlci.GetElementsWithContribution(triangle.as_list))
    
    UpdateMarkers(els_if, mlci.GetElementsWithContribution(triangle.Boundary()))
    UpdateMarkers(els_not_neg, mlci.GetElementsWithContribution(~triangle))

    assert MarkersEqual(els_neg, els_neg2)
    assert MarkersEqual(els_hasneg, els_hasneg2)
    assert MarkersEqual(~els_neg, els_not_neg)
    assert MarkersEqual(els_if, els_hasneg & ~els_neg)

    # --------------------------- Test Integration ----------------------------

    # ---------------------- Test Low-Order Integration -----------------------
    # Codim = 0
    lset_dom_0 = {"levelset": level_sets_p1[0], "domain_type": NEG}
    area0 = Integrate(levelset_domain=lset_dom_0, mesh=mesh, cf=1, order=0)
    assert abs(area0 - 3) < 1e-12

    lset_dom_tri = {"levelset": level_sets_p1, "domain_type": triangle}
    area_tri = Integrate(levelset_domain=lset_dom_tri, mesh=mesh, cf=1, order=0)
    assert abs(area_tri - 0.5) < 1e-12

    lset_dom_tri_i = {"levelset": level_sets_p1, "domain_type": (~triangle)}
    area_tri_i = Integrate(levelset_domain=lset_dom_tri_i,
                           mesh=mesh, cf=1, order=0)
    assert abs(area_tri_i - 3.5) < 1e-12

    # Codim = 1
    lset_dom_side_t = {"levelset": level_sets_p1, "domain_type": (IF, NEG, NEG)}
    length_t = Integrate(levelset_domain=lset_dom_side_t, mesh=mesh, cf=1, order=0)
    assert abs(length_t - 1) < 1e-12

    lset_dom_side_r = {"levelset": level_sets_p1, "domain_type": (NEG, IF, NEG)}
    length_r = Integrate(levelset_domain=lset_dom_side_r, mesh=mesh, cf=1, order=0)
    assert abs(length_r - sqrt(5 / 4)) < 1e-12

    lset_dom_side_l = {"levelset": level_sets_p1, "domain_type": (NEG, NEG, IF)}
    length_l = Integrate(levelset_domain=lset_dom_side_l, mesh=mesh, cf=1, order=0)
    assert abs(length_l - sqrt(5 / 4)) < 1e-12

    lset_dom_per = {"levelset": level_sets_p1,
                    "domain_type": triangle.Boundary()}
    length_per = Integrate(levelset_domain=lset_dom_per, mesh=mesh, cf=1, order=0)
    assert abs(length_per - 1 - sqrt(5)) < 1e-12

    # Codim = 2
    lset_dom_pnt_tl = {"levelset": level_sets_p1, "domain_type": (IF, IF, NEG)}
    point_val_tl = Integrate(levelset_domain=lset_dom_pnt_tl, mesh=mesh, cf=x + y,
                             order=0)
    assert abs(point_val_tl - 1.5) < 1e-12

    lset_dom_pnt_tr = {"levelset": level_sets_p1, "domain_type": (IF, NEG, IF)}
    point_val_tr = Integrate(levelset_domain=lset_dom_pnt_tr, mesh=mesh, cf=x + y,
                             order=0)
    assert abs(point_val_tr - 0.5) < 1e-12

    lset_dom_pnt_b = {"levelset": level_sets_p1, "domain_type": (NEG, IF, IF)}
    point_val_b = Integrate(levelset_domain=lset_dom_pnt_b, mesh=mesh, cf=x + y,
                            order=0)
    assert abs(point_val_b) < 1e-12

    lset_dom_cnrs = {"levelset": level_sets_p1,
                     "domain_type": triangle.Boundary().Boundary()}
    point_val_cnrs = Integrate(levelset_domain=lset_dom_cnrs, mesh=mesh, cf=x + y,
                               order=0)
    assert abs(point_val_cnrs - 2.0) < 1e-12


    del mesh, level_sets, level_sets_p1, mlci, triangle
    del els_neg, els_hasneg, els_if, els_not_neg


def test_2d_ho_integration():
    # -------------------- Test Higher-Order Integration ----------------------
    level_sets = [-y, x - 1, y - 1, -x]
    nr_ls = len(level_sets)

    # ---------------------------- Background Mesh ----------------------------
    geo = SplineGeometry()
    geo.AddRectangle((-0.2, -0.2), (1.2, 1.2),
                     bcs=("bottom", "right", "top", "left"))
    mesh = Mesh(geo.GenerateMesh(maxh=0.4))

    # ------------------------------- LEVELSET --------------------------------
    level_sets_p1 = tuple(GridFunction(H1(mesh, order=1)) for i in range(nr_ls))
    for i, lsetp1 in enumerate(level_sets_p1):
        InterpolateToP1(level_sets[i], lsetp1)

    square = DomainTypeArray([(NEG, NEG, NEG, NEG)])
    lset_dom_square = {"levelset": level_sets_p1, "domain_type": square}

    # ------------------------------- Integrate -------------------------------
    result = Integrate(levelset_domain=lset_dom_square, mesh=mesh, 
                       cf=x * (1 - x) * y * (1 - y), order=4)
    assert abs(result - 1/36) < 1e-12

    del mesh, level_sets, level_sets_p1, square 


def test_2d_overlaps():
    # ---------------------------- Background Mesh ----------------------------
    geo = SplineGeometry()
    geo.AddRectangle((-1.1, -1.1), (1.1, 1.1), bc=1)
    mesh = Mesh(geo.GenerateMesh(maxh=0.02))

    # ------------------------------ Level Sets -------------------------------
    level_sets = [x * x + y * y - 1, -x - 1 / 3, x - 1 / 3, y - 0.5]
    nr_ls = len(level_sets)
    level_sets_p1 = tuple(GridFunction(H1(mesh, order=1)) for i in range(nr_ls))

    for i, lsetp1 in enumerate(level_sets_p1):
        InterpolateToP1(level_sets[i], lsetp1)

    # --------------------------- DomainTypeArrays ----------------------------
    z_disc1 = DomainTypeArray([(NEG, NEG, NEG, POS), (NEG, POS, NEG, POS),
                               (NEG, POS, NEG, NEG), (NEG, NEG, POS, NEG),
                               (NEG, NEG, POS, POS)])
    z_disc2 = DomainTypeArray([(NEG, ANY, ANY, POS), (NEG, POS, ANY, ANY),
                               (NEG, ANY, POS, ANY)])
    part1 = DomainTypeArray((NEG,ANY,ANY,ANY))
    part2 = DomainTypeArray((NEG,NEG,NEG,NEG))
    z_disc3 = part1 & ~part2

    z_disc4 = DomainTypeArray(z_disc3.as_list, level_sets_p1, persistent_compress=False)
    z_disc5 = DomainTypeArray(z_disc3.as_list, level_sets_p1, persistent_compress=True)

    assert z_disc4.lsets == None
    assert z_disc4.persistent_compress == False

    # ----------------------- Test Overlapping Domains ------------------------
    lset_zdisc1 = {"levelset": level_sets_p1, "domain_type": z_disc1}
    lset_zdisc2 = {"levelset": level_sets_p1, "domain_type": z_disc2}
    lset_zdisc3 = {"levelset": level_sets_p1, "domain_type": z_disc3}
    
    area1 = Integrate(lset_zdisc1, 1, mesh, order=0)
    area2 = Integrate(lset_zdisc2, 1, mesh, order=0)
    area3 = Integrate(lset_zdisc3, 1, mesh, order=0)

    assert abs(area1 - area2) < 1e-12
    assert abs(area1 - area3) < 1e-12

    # --------------------------- Test Compression ----------------------------
    z_disc2.Compress(level_sets_p1, persistent=True)
    z_disc3.Compress(level_sets_p1)

    assert z_disc1 == z_disc2
    assert z_disc1 == z_disc3
    assert z_disc1 == z_disc4
    assert z_disc1 == z_disc5

    z_disc2_bnd = z_disc2.Boundary()
    z_disc3_bnd_cmpr = z_disc3.Boundary()
    z_disc3_bnd_cmpr.Compress(level_sets_p1)
    z_disc5_bnd = z_disc5.Boundary()

    assert z_disc2_bnd == z_disc5_bnd
    assert z_disc2_bnd == z_disc3_bnd_cmpr
    
    assert z_disc2_bnd.persistent_compress == True
    for lset1, lset2 in zip(z_disc2_bnd.lsets, level_sets_p1):
        assert lset1 == lset2

    assert z_disc3_bnd_cmpr.persistent_compress == False
    assert z_disc3_bnd_cmpr.lsets == None 

    lset_zdisc2_c = {"levelset": level_sets_p1, "domain_type": z_disc2}
    lset_zdisc3_c = {"levelset": level_sets_p1, "domain_type": z_disc3}
    
    area2_c = Integrate(lset_zdisc2_c, 1, mesh, order=0)
    area3_c = Integrate(lset_zdisc3_c, 1, mesh, order=0)

    assert abs(area1 - area2_c) < 1e-12
    assert abs(area1 - area3_c) < 1e-12

    del mesh, level_sets, level_sets_p1, z_disc1, z_disc2, z_disc3 


def test_multilevelsetcutinfo():
    # ---------------------------- Background Mesh ----------------------------
    geo = SplineGeometry()
    geo.AddRectangle((-1, -1), (1, 1), bc=1)
    mesh = Mesh(geo.GenerateMesh(maxh=0.2))

    # ------------------------------ Test Update ------------------------------
    ba = [BitArray(mesh.ne) for i in range(4)]
    for ba_ in ba:
        ba_.Clear()

    P1 = H1(mesh, order=1)
    lsets = tuple(GridFunction(P1) for i in range(2))

    InterpolateToP1(x + 0.5, lsets[0])
    InterpolateToP1(x - 0.5, lsets[1])

    mlci = MultiLevelsetCutInfo(mesh, lsets)
    ba[0] |= mlci.GetElementsOfType((POS, NEG))

    InterpolateToP1(y + 0.5, lsets[0])
    InterpolateToP1(y - 0.5, lsets[1])    

    ba[1] |= mlci.GetElementsOfType((POS, NEG))
    assert MarkersEqual(ba[0], ba[1])

    mlci.Update(lsets)
    ba[2] |= mlci.GetElementsOfType((POS, NEG))
    assert MarkersEqual(ba[0], ba[2]) == False

    # -------------------------- Test Safety Checks ---------------------------
    with pytest.raises(netgen.libngpy._meshing.NgException):
        mlci2 = MultiLevelsetCutInfo(mesh, (x, y))

    with pytest.raises(netgen.libngpy._meshing.NgException):
        a1 = mlci.GetElementsOfType((NEG, NEG, NEG))
    with pytest.raises(netgen.libngpy._meshing.NgException):
        a2 = mlci.GetElementsOfType([(NEG, NEG, NEG)])
    with pytest.raises(netgen.libngpy._meshing.NgException):
        a3 = mlci.GetElementsWithContribution((NEG, NEG, NEG))
    with pytest.raises(netgen.libngpy._meshing.NgException):
        a4 = mlci.GetElementsWithContribution([(NEG, NEG, NEG)])


# -----------------------------------------------------------------------------
# --------------------------------- 3D TESTS ----------------------------------
# -----------------------------------------------------------------------------
from netgen.csg import *
import collections


def CompList(list1, list2):
    return collections.Counter(list1) == collections.Counter(list2)

def test_3d_mlset():
    # ---------------------------- Background Mesh ----------------------------
    geo = CSGeometry()
    geo.Add(OrthoBrick(Pnt(-0.8, -0.8, -0.8), Pnt(0.8, 0.8, 0.8)))
    mesh = Mesh(geo.GenerateMesh(maxh=0.5))

    # ------------------------------ Level Sets -------------------------------
    level_sets = [x - 0.5, x + 0.5, y - 0.5, y + 0.5, z - 0.5, z + 0.5]
    nr_ls = len(level_sets)
    level_sets_p1 = tuple(GridFunction(H1(mesh, order=1)) for i in range(nr_ls))
    for i, lset_p1 in enumerate(level_sets_p1):
        InterpolateToP1(level_sets[i], lset_p1)

    # ---------------------------- DomainTypeArray ----------------------------
    cube = DomainTypeArray(dtlist=[(NEG, POS, NEG, POS, NEG, POS)])
    sides = []
    for i in range(nr_ls):
        domain = [NEG, POS, NEG, POS, NEG, POS]
        domain[i] = IF
        sides.append(DomainTypeArray(tuple(domain)))

    target_list = [dtt for side in sides for dtt in side.as_list]
    assert CompList(cube.Boundary().as_list, target_list)

    # ------------------------------- Cut-Info --------------------------------
    mlci = MultiLevelsetCutInfo(mesh, level_sets_p1)

    # ------------------------- Element Marker Tests --------------------------
    els_neg, els_hasneg = BitArray(mesh.ne), BitArray(mesh.ne)
    els_not_neg, els_if = BitArray(mesh.ne), BitArray(mesh.ne)

    UpdateMarkers(els_neg, mlci.GetElementsOfType(cube))
    UpdateMarkers(els_hasneg, mlci.GetElementsWithContribution(cube))
    UpdateMarkers(els_if, mlci.GetElementsWithContribution(cube.Boundary()))
    UpdateMarkers(els_not_neg, mlci.GetElementsWithContribution(~cube))

    assert MarkersEqual(~els_neg, els_not_neg)
    assert MarkersEqual(els_if, els_hasneg & ~els_neg)

    els_sides = BitArray(mesh.ne)
    els_sides[:] = False
    for side in sides:
        els_sides |= mlci.GetElementsWithContribution(side)

    assert MarkersEqual(els_sides, els_if)

    # --------------------------- Test Integration ----------------------------

    # Codim = 0
    lset_dom_cube = {"levelset": level_sets_p1, "domain_type": cube}
    volume_cube = Integrate(levelset_domain=lset_dom_cube,
                            mesh=mesh, cf=1, order=0)
    assert abs(volume_cube - 1) < 1e-12

    lset_dom_cube_inv = {"levelset": level_sets_p1, "domain_type": ~cube}
    volume_cube_inv = Integrate(levelset_domain=lset_dom_cube_inv, mesh=mesh, cf=1,
                                order=0)
    assert abs(volume_cube_inv - 3.096) < 1e-12

    # Codim = 1
    for side in sides:
        lset_dom_side = {"levelset": level_sets_p1, "domain_type": side}
        area_side = Integrate(levelset_domain=lset_dom_side, mesh=mesh, cf=1,
                              order=0)
        assert abs(area_side - 1) < 1e-12

    lset_dom_surface = {"levelset": level_sets_p1,
                        "domain_type": cube.Boundary()}
    area_surface = Integrate(levelset_domain=lset_dom_surface, mesh=mesh, cf=1,
                             order=0)
    assert abs(area_surface - 6) < 1e-12


    # Codim = 2
    edges = [(IF, POS, IF, POS, NEG, POS), (IF, POS, NEG, IF, NEG, POS),
             (IF, POS, NEG, POS, IF, POS), (IF, POS, NEG, POS, NEG, IF),
             (NEG, IF, IF, POS, NEG, POS), (NEG, IF, NEG, IF, NEG, POS),
             (NEG, IF, NEG, POS, IF, POS), (NEG, IF, NEG, POS, NEG, IF),
             (NEG, POS, IF, POS, IF, POS), (NEG, POS, NEG, IF, IF, POS),
             (NEG, POS, IF, POS, NEG, IF), (NEG, POS, NEG, IF, NEG, IF)]

    for edge in edges:
        lset_dom_edge = {"levelset": level_sets_p1, "domain_type": edge}
        len_edge = Integrate(levelset_domain=lset_dom_edge, mesh=mesh, cf=1,
                              order=0)
        assert abs(len_edge - 1) < 1e-12

    lset_dom_edges = {"levelset": level_sets_p1,
                        "domain_type": cube.Boundary().Boundary()}
    area_surface = Integrate(levelset_domain=lset_dom_edges, mesh=mesh, cf=1,
                             order=0)
    assert abs(area_surface - 12) < 1e-12 


    # Codim = 3
    point_domains = {"ppp": (IF, POS, IF, POS, IF, POS),
                     "ppn": (IF, POS, IF, POS, NEG, IF),
                     "pnp": (IF, POS, NEG, IF, IF, POS),
                     "npp": (NEG, IF, IF, POS, IF, POS),
                     "pnn": (IF, POS, NEG, IF, NEG, IF),
                     "npn": (NEG, IF, IF, POS, NEG, IF),
                     "nnp": (NEG, IF, NEG, IF, IF, POS),
                     "nnn": (NEG, IF, NEG, IF, NEG, IF)}

    vals = {"ppp": 1.5, "ppn": 0.5, "pnp": 0.5, "npp": 0.5, "pnn": -0.5,
            "npn": -0.5, "nnp": -0.5, "nnn": -1.5}

    for i, (key, domain) in enumerate(point_domains.items()):
        lset_pnt = {"levelset": level_sets_p1, "domain_type": domain}
        point = Integrate(levelset_domain=lset_pnt, mesh=mesh, cf=x + y + z,
                          order=0)
        assert abs(point - vals[key]) < 1e-12


def test_3d_codim2_cross():

    # ---------------------------- Background Mesh ----------------------------
    geo = CSGeometry()
    geo.Add(OrthoBrick(Pnt(-0.8,-0.8,-0.8), Pnt(0.8,0.8,0.8)))
    mesh = Mesh(geo.GenerateMesh(maxh=0.8))

    # ------------------------------ Level Sets -------------------------------
    level_sets = (x - 0.5, x + 0.5, x - y, z - 0)
    nr_ls = len(level_sets)
    level_sets_p1 = tuple(GridFunction(H1(mesh,order=1)) for i in range(nr_ls))

    for i, lset_p1 in enumerate(level_sets_p1):
        InterpolateToP1(level_sets[i], lset_p1)

    # ---------------------------- DomainTypeArray ----------------------------
    line = DomainTypeArray(dtlist=[(NEG, POS, IF, IF)])

    # --------------------------- Test Integration ----------------------------
    length = Integrate(levelset_domain={"levelset": level_sets_p1,
                                        "domain_type": line},
                       mesh=mesh, cf=1, order=0)
    assert abs(length - sqrt(2)) < 1e-12

    del level_sets, level_sets_p1, nr_ls, line


    # ------------------------------ SECOND TEST ------------------------------

    # ------------------------------ Level Sets -------------------------------
    level_sets = (x - 0.5, - x - 0.5, z - y, x - z, x + z)
    nr_ls = len(level_sets)
    level_sets_p1 = tuple(GridFunction(H1(mesh,order=1)) for i in range(nr_ls))

    for i, lset_p1 in enumerate(level_sets_p1):
        InterpolateToP1(level_sets[i], lset_p1)

    # ---------------------------- DomainTypeArray ----------------------------
    cross = DomainTypeArray([(NEG, NEG, IF, IF, ANY), (NEG, NEG, IF, ANY, IF)])

    # --------------------------- Test Integration ----------------------------
    length = Integrate(levelset_domain={"levelset": level_sets_p1,
                                        "domain_type": cross},
                       mesh=mesh, cf=1, order=0)
    assert abs(length - 2* sqrt(3)) < 1e-12

    del level_sets, level_sets_p1, nr_ls, cross


    # ------------------------------ THIRD TEST -------------------------------

    # ------------------------------ Level Sets -------------------------------
    level_sets = (x - 0.5, - x - 0.5, - z - 0.5, z + y, x + y, z + y + 0.1)
    nr_ls = len(level_sets)
    level_sets_p1 = tuple(GridFunction(H1(mesh,order=1)) for i in range(nr_ls))

    for i, lset_p1 in enumerate(level_sets_p1):
        InterpolateToP1(level_sets[i], lset_p1)

    # ---------------------------- DomainTypeArray ----------------------------
    lines = DomainTypeArray([(NEG, NEG, NEG, IF, IF, ANY), 
                             (NEG, NEG, NEG, ANY, IF, IF)])

    # --------------------------- Test Integration ----------------------------
    length = Integrate(levelset_domain={"levelset": level_sets_p1,
                                        "domain_type": lines},
                       mesh=mesh, cf=1, order=0)
    assert abs(length - 1.9*sqrt(3)) < 1e-12


# -----------------------------------------------------------------------------
# --------------------------- RUN TESTS SEPARATELY ----------------------------
# -----------------------------------------------------------------------------
if __name__ == "__main__":
    test_2d_mlci_and_lo_integration()
    test_multilevelsetcutinfo()
    test_2d_ho_integration()
    test_2d_overlaps()
    test_3d_mlset()
    test_3d_codim2_cross()


from os.path import dirname, isdir, join
import os
import re
import subprocess
def get_version():
    """
    Gets the current version number.
    If in a git repository, it is the current git tag.
    Otherwise it is the one contained in the PKG-INFO file.
    """
    version_re = re.compile('^Version: (.+)$', re.M)
    d = dirname(__file__)+'/../..'

    if isdir(join(d, '.git')):
        # Get the version using "git describe".
        cmd = 'git describe --always --tags --match v[0-9]*'.split()
        try:
            version = subprocess.check_output(cmd).decode().strip()
        except subprocess.CalledProcessError:
            print('Unable to get version number from git tags')
            exit(1)

        # PEP 386 compatibility
        if '-' in version:
            version = '.post'.join(version.split('-')[:2])

        # Don't declare a version "dirty" merely because a time stamp has
        # changed. If it is dirty, append a ".dev1" suffix to indicate a
        # development revision after the release.
        with open(os.devnull, 'w') as fd_devnull:
            subprocess.call(['git', 'status'],
                            stdout=fd_devnull, stderr=fd_devnull)

        cmd = 'git diff-index --name-only HEAD'.split()
        try:
            dirty = subprocess.check_output(cmd).decode().strip()
        except subprocess.CalledProcessError:
            print('Unable to get git index status')
            exit(1)

        # if dirty != '':
            # version += '.dev1'

        # strip the v for pypi
        version = version[1:]

    else:
        # Extract the version from the PKG-INFO file.
        with open(join(d, 'PKG-INFO')) as f:
            version = version_re.search(f.read()).group(1)

    return version


# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'ngsxfem'
copyright = '2022, Christoph Lehrenfeld and the ngsxfem developer'
author = 'Christoph Lehrenfeld, Fabian Heimann, Henry von Wahl, Janosch Preuß and the ngsxfem community'

# The full version, including alpha/beta/rc tags
release = get_version()

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx_mdinclude",
    "jupyter_sphinx",
    "sphinx.ext.autodoc",
    "myst_nb",
    ]

templates_path = ['_templates']
exclude_patterns = []



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

master_doc = "contents"

source_suffix = ['.rst', '.md']

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']
html_css_files = ['bullets.css']

html_sidebars = {
   'index': [
        'about.html',
        'localtoc.html',
        # 'relations.html',
       ],
   '**': ['about.html',
          'localtoc.html',
       ],
}

def setup(app):
    app.add_css_file("custom.css")


# enable latex rendering in myst-nb
myst_enable_extensions = ["dollarmath", "amsmath"]


"""
	Description:
	------------
	Sharp cube defined by three levelsets.
"""

# Import libraries
from netgen.csg import *
from ngsolve import *
from xfem import *
from xfem.mlset import *

# Mesh
geo = CSGeometry()
geo.Add(OrthoBrick(Pnt(-0.8,-0.8,-0.8), Pnt(0.8,0.8,0.8)))
mesh = Mesh(geo.GenerateMesh(maxh=0.75))

# Level set
V = H1(mesh,order=1)
lsetp1_1, lsetp1_2, lsetp1_3 = GridFunction(V), GridFunction(V), GridFunction(V)

def cf_abs(x):
	return IfPos(x,x,-x)
InterpolateToP1( cf_abs(x) - 0.5, lsetp1_1)
InterpolateToP1( cf_abs(y) - 0.5, lsetp1_2)
InterpolateToP1( cf_abs(z) - 0.5, lsetp1_3)


Draw(lsetp1_1,mesh,"lsetp1_1")
Draw(lsetp1_2,mesh,"lsetp1_2")
Draw(lsetp1_3,mesh,"lsetp1_3")

# Integrate
cube = DomainTypeArray(dtlist=[(NEG,NEG,NEG)])

volume = Integrate(levelset_domain={"levelset": [lsetp1_1, lsetp1_2, lsetp1_3], 
                                    "domain_type": cube.as_list},
                   mesh=mesh, cf=1, order=0)


print("volume = {:10.8f}".format(volume))            
print("volume error = {:4.3e}".format(abs(volume - 1)))  
assert abs(volume - 1) < 1e-12

surface_area = Integrate(levelset_domain={"levelset": [lsetp1_1, lsetp1_2, lsetp1_3], 
                                          "domain_type": cube.Boundary().as_list},
                         mesh=mesh, cf=1, order=0)

print("surface_area = {:10.8f}".format(surface_area))            
print("surface_area error = {:4.3e}".format(abs(surface_area - 6)))  
assert abs(surface_area - 6) < 1e-12


from ngsolve import *
from time import sleep
from netgen.geom2d import unit_square
from netgen.geom2d import SplineGeometry
from netgen.meshing import MeshingParameters

from ngsolve.internal import *
from xfem import *
from numpy import pi

from xfem.lset_spacetime import *

# geometry
# Geometry 
square = SplineGeometry()
square.AddRectangle([0,0],[2,2],bc=1)
maxh = 0.1
mesh = Mesh (square.GenerateMesh(maxh=maxh, quad_dominated=False))

n_ref = 0
for i in range(n_ref):
    mesh.Refine()

h = specialcf.mesh_size
n_outer = specialcf.normal(mesh.dim)



# FE Spaces
k_t = 2
k_s = 2
fes1 = H1(mesh, order=k_s,dirichlet=[1])
fes_lset_slice = H1(mesh, order=1, dirichlet=[])
fes_dfm_slice = H1(mesh, order=k_s, dim=mesh.dim)

lset_order_time = 2
time_order = 4
tfe = ScalarTimeFE(k_t) 

st_fes = SpaceTimeFESpace(fes1,tfe,dirichlet=[1])
VhG_ic = FESpace([fes1,fes1])
VhG = FESpace([st_fes,st_fes])


tend = 0.5
delta_t = 1/16
tnew = 0

told = Parameter(0)
tref = ReferenceTimeVariable()
t = told + delta_t*tref

lset_adap_st = LevelSetMeshAdaptation_Spacetime(mesh, order_space = k_s, order_time = lset_order_time,
                                         threshold=0.1, discontinuous_qn=True)

# Problem data
rho =  CoefficientFunction((1/(4*pi))*sin(2*pi*t))
d_rho = CoefficientFunction(0.5*cos(2*pi*t))

#rho =  CoefficientFunction(0)
#d_rho = CoefficientFunction(0)

alpha_neg = 10.0
alpha_pos = 20.0
alpha = [alpha_neg,alpha_pos]
beta_neg = 2.0
beta_pos = 1.0
beta = [beta_neg,beta_pos]

r0 = 1/3
w = CoefficientFunction((d_rho,0)) # convection
levelset= CoefficientFunction(sqrt((x-0.5-rho)*(x-0.5-rho)+(y-1)*(y-1)) -r0)
a_const = 1.1569
b_const = -8.1621
k_const = 1

r_lset = CoefficientFunction(sqrt((x-0.5-rho)*(x-0.5-rho) + (y-1)*(y-1)))
f_pos = CoefficientFunction(   pi*cos(pi*t)*cos(pi*r_lset) 
                             + alpha_pos*sin(pi*t)*pi*pi*cos(pi*r_lset)
                             + alpha_pos*pi*sin(pi*t)*sin(pi*r_lset)/r_lset )
f_neg = ( k_const*pi*cos(k_const*pi*t)*(a_const+b_const*((x-0.5-rho)*(x-0.5-rho)+(y-1)*(y-1))) 
         - alpha_neg*sin(k_const*pi*t)*2*b_const
         - alpha_neg*sin(k_const*pi*t)*2*b_const   )
coeff_f = [f_neg,f_pos]


# for monitoring the error
exact = [sin(k_const*pi*t)*(a_const+b_const*((x-0.5-rho)*(x-0.5-rho)+(y-1)*(y-1)) ),
         sin(k_const*pi*t)*cos(pi*sqrt((x-0.5-rho)*(x-0.5-rho)+(y-1)*(y-1)))]

u0 = GridFunction(VhG)
u0_ic = GridFunction(VhG_ic)
#Draw(u0_ic,mesh,"u")

# Definition of Trial/Test-functions
u = VhG.TrialFunction()
v = VhG.TestFunction()

gradu = [grad(ui) for ui in u]
gradv = [grad(vi) for vi in v]


lset_top = GridFunction(fes_lset_slice)
lset_bottom = GridFunction(fes_lset_slice)
dfm_top = GridFunction(fes_dfm_slice)

t_old = 0
coef_u0 = [CoefficientFunction(0),CoefficientFunction(0)]
u0_ic.components[0].Set(coef_u0[0]) # set initial condition
u0_ic.components[1].Set(coef_u0[1]) # set initial condition

stab = 10*(alpha[0]+alpha[1])*(k_s+1)*(k_s+1)/h
stab_outer = 10*(k_s+1)*(k_s+1)/h


while tend - t_old > delta_t/2:
    
    dfm = lset_adap_st.CalcDeformation(levelset,told,t_old,delta_t,calc_kappa = True) 
    dfm_top.vec[:] = dfm.vec[lset_order_time*lset_adap_st.ndof_node : (lset_order_time+1)*lset_adap_st.ndof_node]
    lset_p1 = lset_adap_st.lset_p1  
    active_dofs = BitArray(VhG.FreeDofs())
   
    hasneg_spacetime = lset_adap_st.hasneg_spacetime
    hasneg_spacetime |= lset_adap_st.hasif_spacetime
    haspos_spacetime = lset_adap_st.haspos_spacetime
    haspos_spacetime |= lset_adap_st.hasif_spacetime
    
    active_dofs &= CompoundBitArray([GetDofsOfElements(st_fes,hasneg_spacetime),GetDofsOfElements(st_fes,haspos_spacetime)])
    #print(active_dofs) 
    
    lset_neg = { "levelset" : lset_p1, "domain_type" : NEG, "subdivlvl" : 0}
    lset_pos = { "levelset" : lset_p1, "domain_type" : POS, "subdivlvl" : 0}
    lset_if = { "levelset" : lset_p1, "domain_type" : IF, "subdivlvl" : 0}
    # Put this into LevelSetMeshAdaptation_Spacetime later on
    lset_bottom.vec[:] = lset_p1.vec[0:fes_lset_slice.ndof]
    lset_neg_bottom = { "levelset" : lset_bottom, "domain_type" : NEG, "subdivlvl" : 0}
    lset_pos_bottom = { "levelset" : lset_bottom, "domain_type" : POS, "subdivlvl" : 0}
    lset_top.vec[:] = lset_p1.vec[lset_order_time*fes_lset_slice.ndof : (lset_order_time+1)*fes_lset_slice.ndof]
    lset_neg_top = { "levelset" : lset_top, "domain_type" : NEG, "subdivlvl" : 0}
    lset_pos_top = { "levelset" : lset_top, "domain_type" : POS, "subdivlvl" : 0}
    
    n_lset_p1 = 1.0/grad(lset_adap_st.lset_p1).Norm() * grad(lset_adap_st.lset_p1)
    
    
    kappas = {"interp" : (lset_adap_st.kappa,1-lset_adap_st.kappa), 
              "st_CutInfo" : (CutRatioGF(lset_adap_st.ci),1-CutRatioGF(lset_adap_st.ci)) }
    kappa = kappas["st_CutInfo"]
    
#    time_quad = lset_adap_st.v_ho_st.TimeFE_nodes().NumPy()
#    times = [t_old + delta_t * xi for xi in time_quad]
#    for tau in times:  
#        lset_adap_st.v_kappa.SetTime(tau)
#        Draw(kappa[0],mesh,"kappa")
#        input("Kappa at t = {0}".format(tau))
#    lset_adap_st.v_kappa.SetOverrideTime(False)
         
    average_flux_u = sum( [-kappa[i]*alpha[i]*gradu[i]*n_lset_p1 for i in [0,1]] )
    average_flux_v = sum( [-kappa[i]*alpha[i]*gradv[i]*n_lset_p1 for i in [0,1]] )
    betajump_u = beta[0]*u[0]-beta[1]*u[1]
    betajump_v = beta[0]*v[0]-beta[1]*v[1]
    
     
    a = BilinearForm(VhG,symmetric=False)

    a += SymbolicBFI(levelset_domain = lset_neg, form =  -u[0]*dt(v[0],0)*beta[0]
                                                      + delta_t*alpha[0]*beta[0]*gradu[0]*gradv[0]
                                                      - delta_t*beta[0]*u[0]*gradv[0]*w, time_order=time_order)    
                                                      
    a += SymbolicBFI(levelset_domain = lset_pos, form = -u[1]*dt(v[1],1)*beta[1]  
                                                      + delta_t*alpha[1]*beta[1]*gradu[1]*gradv[1]
                                                      - delta_t*beta[1]*u[1]*gradv[1]*w, time_order=time_order)
 
    
    a += SymbolicBFI(levelset_domain = lset_neg_top, form = beta[0]*fix_t(u[0],1,0)*fix_t(v[0],1,0) )
    a += SymbolicBFI(levelset_domain = lset_pos_top, form = beta[1]*fix_t(u[1],1,1)*fix_t(v[1],1,1) )
    
    

    a += SymbolicBFI(levelset_domain = lset_if, form = delta_t*( average_flux_u*betajump_v   
                                                              + betajump_u*average_flux_v 
                                                            + stab*betajump_u*betajump_v), time_order=time_order )

                                                               
                
    f = LinearForm(VhG)
    f += SymbolicLFI(levelset_domain = lset_neg, form = delta_t*beta[0]*coeff_f[0]*v[0], time_order=time_order)
    f += SymbolicLFI(levelset_domain = lset_pos, form = delta_t*beta[1]*coeff_f[1]*v[1], time_order=time_order)
    f += SymbolicLFI(levelset_domain = lset_neg_bottom,form = beta[0]*u0_ic.components[0]*fix_t(v[0],0,0) )
    f += SymbolicLFI(levelset_domain = lset_pos_bottom,form = beta[1]*u0_ic.components[1]*fix_t(v[1],0,1) )
     
    
    mesh.SetDeformation(dfm)
    a.Assemble()
    f.Assemble()

    mesh.UnsetDeformation()
    
    # Setting boundary values (temporary solution)
    u0.vec[:] = 0
   
    times = [t_old + delta_t * xi for xi in st_fes.TimeFE_nodes()]
    u_bnd = GridFunction(fes1)
    for i,ti in enumerate(times):
        rho_ti = (1/(4*pi))*sin(2*pi*ti)
        print(ti)
        u_bnd.Set(sin(k_const*pi*ti)*cos(pi*sqrt((x-0.5-rho_ti)*(x-0.5-rho_ti)+(y-1)*(y-1))),BND)
        u0.components[1].vec[i*fes1.ndof : (i+1)*fes1.ndof] = u_bnd.vec[:] 
    
    #u_bnd.vec[:] = u0.components[1].vec[0 : fes1.ndof]   
    #Draw(sqrt((u_bnd-exact[1])*(u_bnd-exact[1])),mesh,"errorB")
    #input("Stop")
    
           
    res = f.vec.CreateVector()
    res.data = f.vec - a.mat * u0.vec
    u0.vec.data += a.mat.Inverse(active_dofs,"") * res
           
    
       
    # exploiting the nodal property of the time fe:b_const
    for i in range(2): 
        u0_ic.components[i].vec[:] = u0.components[i].vec[k_t*fes1.ndof : (k_t+1)*fes1.ndof]
   
    
    
    t_old = t_old + delta_t
    told.Set(t_old)
    #u0_ic.components[1].Set(exact[1])
    
    mesh.SetDeformation(dfm_top)
    err_sqr_coefs = [ (u0_ic.components[i] - exact[i])*(u0_ic.components[i] - exact[i]) for i in [0,1] ]
    l2error = sqrt( Integrate( levelset_domain=lset_neg_top, cf=err_sqr_coefs[0], mesh=mesh,
                           order=2*k_s, heapsize=1000000)
                + Integrate(levelset_domain=lset_pos_top, cf=err_sqr_coefs[1], mesh=mesh,
                            order=2*k_s, heapsize=1000000)) 
    mesh.UnsetDeformation()
    print("t = {0}, l2error = {1}".format(t_old,l2error))
    
    Draw(IfPos(lset_top,u0_ic.components[1],u0_ic.components[0]),mesh,"u")
    #Draw(IfPos(lset_top,sqrt(err_sqr_coefs[1]),sqrt(err_sqr_coefs[0])),mesh,"error")
    #Draw(IfPos(lset_top,exact[1],exact[0]),mesh,"exact")
    #input("Continue")
    #Redraw(blocking=True)
       



"""
In this example we consider a scalar time-dependent convection-diffusion
problem on a moving domain. The discretisation uses a level set based
geometry description, a cut finite element method for the spatial
discretisation and utilised ghost-penalty stabilisation to realise a
discrete extension of the solution into an extension strip. With this
extension, a method of lines approximation of the time-derivative is
possible to create an Eulerian time-stepping method. For the numerical
analysis of this method, see [1]. In contrast to [1], this file uses
an isoparametric mapping approach to realise a higher-order geometry
approximation.

Used Features:
--------------
* Level set geometry description and element markings, see the 'basic'
  jupyter tutorial.

* Integrators with respect to level set geometry, see the 'cutfem'
  jupyter tutorial.

* Higher-order geometry approximation, see the 'cutfem' jupyter
  tutorial.

* Ghost-penalty stabilisation, see the 'spacetime' jupyter tutorial.

Literature:
-----------
[1] C. Lehrenfeld, M. Olshanskii, An Eulerian finite element method for
    PDEs in time-dependent domains. ESAIM M2AN 53(2):585-614, 2019.
"""

# ------------------------------ LOAD LIBRARIES -------------------------------
from netgen.geom2d import SplineGeometry
from ngsolve import *
from xfem import *
from xfem.lsetcurv import *

from math import pi, ceil

ngsglobals.msg_level = 2
SetNumThreads(4)

# -------------------------------- PARAMETERS ---------------------------------
# Finite element space order
k = 1
# Initial mesh size and time-step size
h0, t0 = 0.2, 0.1
# Number of mesh and time-step bisections
Lx, Lt = 3, 3

# Background geometry corners
lowerleft, upperright = [-0.7, -0.7], [0.9, 0.7]

# Time parameters
dt = t0 * 0.5**Lt
T_end = 0.4
t = Parameter(0.0)

# Problem and discretisation parameters
nu = 1e-5
c_gamma = 1

# Sparse direct solver
inverse = ""

# Problem Data
# Initial condition and right-hand side
rho = CoefficientFunction(1 / pi * sin(2 * pi * t))

r0 = 0.5
r1 = pi / (2 * r0)

u0 = CoefficientFunction(cos(r1 * sqrt(x**2 + y**2))**2)
u_ex = CoefficientFunction(cos(r1 * sqrt((x - rho)**2 + y**2))**2)

grad_u_ex = CoefficientFunction((-pi * sin(pi / r0 * sqrt((x - rho)**2 + y**2))
                                 * (x - rho) / sqrt((x - rho)**2 + y**2),
                                 - pi * sin(pi / r0 *
                                            sqrt((x - rho)**2 + y**2))
                                 * y / sqrt((x - rho)**2 + y**2)))
# rhs from py_tutorial
rhs = nu * CoefficientFunction(-(pi / r0) * r1
                               * (sin(r1 * sqrt((x - rho)**2 + y**2))
                                  * sin(r1 * sqrt((x - rho)**2 + y**2))
                                  - cos(r1 * sqrt((x - rho)**2 + y**2))
                                  * cos(r1 * sqrt((x - rho)**2 + y**2)))
                               + (pi / r0)
                               * cos(r1 * sqrt((x - rho)**2 + y**2))
                               * sin(r1 * sqrt((x - rho)**2 + y**2))
                               * (1 / sqrt((x - rho)**2 + y**2)))

# Convection Field
w = CoefficientFunction((2 * cos(2 * pi * t), 0))
div_w = 0.0
velmax = 2


def Rho(t):
    return CoefficientFunction(1 / pi * sin(2 * pi * t))


# Level set
def levelset_func(t):
    return sqrt((x - Rho(t))**2 + y**2) - r0


# ----------------------------------- MAIN ------------------------------------
background_domain = SplineGeometry()
background_domain.AddRectangle(lowerleft, upperright, bc=1)
ngmesh = background_domain.GenerateMesh(maxh=h0, quad_dominated=False)
for i in range(Lx):
    ngmesh.Refine()
mesh = Mesh(ngmesh)
h_max = h0 * 0.5**Lx

# FE Space
V = H1(mesh, order=k, dgjumps=True)
gfu = GridFunction(V)


# Higher order discrete level set approximation
lsetmeshadap = LevelSetMeshAdaptation(mesh, order=k, threshold=0.1,
                                      discontinuous_qn=True)
deformation = lsetmeshadap.deform
lsetp1 = lsetmeshadap.lset_p1

# Cut-Info classes for element marking
ci_main = CutInfo(mesh)
ci_inner = CutInfo(mesh)
ci_outer = CutInfo(mesh)

# Element and facet markers
els_hasneg = ci_main.GetElementsOfType(HASNEG)
els_outer = ci_outer.GetElementsOfType(HASNEG)
els_inner = ci_inner.GetElementsOfType(NEG)

els_ring = BitArray(mesh.ne)
facets_ring = BitArray(mesh.nedge)
els_outer_old, els_test = BitArray(mesh.ne), BitArray(mesh.ne)

# Discretisation
delta = 1 * dt * velmax
K_tilde = int(ceil(delta / h_max))
gamma_s = c_gamma * K_tilde

u, v = V.TnT()
h = specialcf.mesh_size

dx = dCut(levelset=lsetp1, domain_type=NEG, definedonelements=els_hasneg,
          deformation=deformation)
dw = dFacetPatch(definedonelements=facets_ring, deformation=deformation)

# Bilinear and linear forms
a = RestrictedBilinearForm(V, element_restriction=els_outer,
                           facet_restriction=facets_ring, check_unused=False)
a += (1 / dt) * u * v * dx
a += nu * InnerProduct(Grad(u), Grad(v)) * dx
a += (InnerProduct(w, grad(u)) * v + div_w * u * v) * dx
a += gamma_s * (1 / h**2) * (u - u.Other()) * (v - v.Other()) * dw

f = LinearForm(V)
f += rhs * v * dx
f += (1 / dt) * gfu * v * dx


# Error computation
errors_L2, errors_H1 = [], []
dx_2k = dx.order(2 * k)


def CompErrs():
    l2 = sqrt(Integrate((gfu - u_ex)**2 * dx_2k, mesh))
    h1 = sqrt(Integrate((grad(gfu) - grad_u_ex)**2 * dx_2k, mesh))

    errors_L2.append(l2)
    errors_H1.append(h1)

    return l2


# Project the solution defined with respect to the last mesh deformation
# onto the the mesh with the current mesh deformation.
lsetmeshadap.ProjectOnUpdate(gfu)

# Time stepping loop
gfu.Set(u0)
els_outer_old.Set()

Draw(gfu, mesh, "gfu")
Draw(gfu - u_ex, mesh, "l2err")

for it in range(1, int(T_end / dt + 0.5) + 1):
    t.Set(it * dt)
    # Update mesh deformation (Note: this updates gfu)
    deformation = lsetmeshadap.CalcDeformation(levelset_func(t))

    # Update Element markers: extension facets. Updating the cut-info
    # classes automatically updates the marker BitArrays.
    InterpolateToP1(levelset_func(t) - delta, lsetp1)
    ci_outer.Update(lsetp1)
    InterpolateToP1(levelset_func(t) + delta, lsetp1)
    ci_inner.Update(lsetp1)

    els_ring.Clear(), facets_ring.Clear()
    els_ring |= els_outer & ~els_inner
    facets_ring |= GetFacetsWithNeighborTypes(mesh, a=els_outer, b=els_ring)

    # Update Element markers: Domain elements
    InterpolateToP1(levelset_func(t), lsetp1)
    ci_main.Update(lsetp1)

    active_dofs = GetDofsOfElements(V, els_outer)

    # Check element history for method of lines time-derivative approx.
    els_test[:] = els_hasneg & ~els_outer_old
    assert sum(els_test) == 0, 'Some active elements do not have a history'

    els_outer_old[:] = els_outer

    # Update Linear System
    a.Assemble(reallocate=True)
    f.Assemble()
    inv = a.mat.Inverse(active_dofs, inverse=inverse)
    # Solve Problem
    gfu.vec.data = inv * f.vec

    # Compute Errors
    l2err = CompErrs()

    print(f"Lx = {Lx}, dt = {dt:8.6f}, t = {t.Get():8.6f}, ", end="")
    print(f"error(L2) = {l2err:4.2e}, ", end="")
    print(f"active_els = {sum(els_outer)}, K = {K_tilde}")
    Redraw(blocking=True)


# Post Processing
err_l2l2 = sqrt(dt * sum([e**2 for e in errors_L2]))
err_l2h1 = sqrt(dt * sum([e**2 for e in errors_H1]))
# err_l1l2 = sum([e*dt for e in errors_L2])
# err_l1h1 = sum([e*dt for e in errors_H1])

print("\n--------------------------------------------------------")
print("  L2(0,T;L2) Error = {:6.4e}".format(err_l2l2))
print("Linf(0,T;L2) Error = {:6.4e}".format(max(errors_L2)))
print("  L2(0,T;H1) Error = {:6.4e}".format(err_l2h1))



'''
These test-cases compare the results of SIMD-enabled and
  SIMD-disabled evaluation of e.g. .Apply(), .Assemble(), 
  .AssembleLinearized(). We cannot guarantee that SIMD-enabled
  evaluates faster than without, therefore this is not checked; 
  but can be printed with pytest -s 'filename.py'.
last edited: 08-10-2023 (DD-MM-YYYY)
'''

from xfem import ngsxfemglobals
from timeit import default_timer as timer
import pytest
from xfem import *

from netgen.geom2d import SplineGeometry
from ngsolve import *
from ngsolve.internal import *
from xfem import *
from xfem.lsetcurv import *
import scipy.sparse as sp

from ngsolve.meshes import *
from netgen.csg import *
from math import pi
import netgen.meshing as ngm
from xfem.lset_spacetime import *

import numpy as np


@pytest.mark.parametrize("maxh", [2**(-k) for k in range(5)])
@pytest.mark.parametrize("order", [4])
def test_lf_blf(maxh, order):
    print()
    print("test for CalcElementMatrixAdd(symboliccutbfi and CalcElementVector(symboliccutlfi) with order=" + str(order) + " and maxh=" + str(maxh))
    square = SplineGeometry()
    square.AddRectangle((-1, -1), (1, 1), bc=1)
    ngmesh = square.GenerateMesh(maxh=maxh, quad_dominated=False)
    mesh = Mesh(ngmesh)

    # Manufactured exact solution for monitoring the error
    r2 = 3 / 4  # outer radius
    r1 = 1 / 4  # inner radius
    rc = (r1 + r2) / 2.0
    rr = (r2 - r1) / 2.0
    r = sqrt(x**2 + y**2)
    levelset = IfPos(r - rc, r - rc - rr, rc - r - rr)

    exact = (20 * (r2 - sqrt(x**2 + y**2)) *
             (sqrt(x**2 + y**2) - r1)).Compile()
    coeff_f = - (exact.Diff(x).Diff(x) + exact.Diff(y).Diff(y)).Compile()

    # Higher order level set approximation
    lsetmeshadap = LevelSetMeshAdaptation(mesh, order=order, threshold=0.1,
                                          discontinuous_qn=True)
    deformation = lsetmeshadap.CalcDeformation(levelset)
    lsetp1 = lsetmeshadap.lset_p1

    # Element, facet and dof marking w.r.t. boundary approximation with lsetp1:
    ci = CutInfo(mesh, lsetp1)
    hasneg = ci.GetElementsOfType(HASNEG)

    Vhbase = H1(mesh, order=order, dirichlet=[], dgjumps=True, dim=mesh.dim)
    Vh = Restrict(Vhbase, hasneg)

    u, v = Vh.TrialFunction(), Vh.TestFunction()

    # integration domains:
    dx = dCut(lsetp1, NEG, definedonelements=hasneg, deformation=deformation)

    # R.h.s. term:
    f = LinearForm(Vh)
    f += CF((coeff_f,sin(x))) * v * dx
    g = LinearForm(Vh)
    g += CF((coeff_f,sin(x))) * v * dx
    a = BilinearForm(Vh)
    a += (u) * (v) * dx
    b = BilinearForm(Vh)
    b += (u) * (v) * dx
    t_normal = 0
    t_blf_normal = 0
    t_simd = 0
    t_blf_simd = 0
    n=10
    for k in range(n):
        ngsxfemglobals.SwitchSIMD(False)
        ########################
        # Time Linear Form, no simd
        start = timer()

        f.Assemble()
        end = timer()
        t_normal += end-start
        #########################
        # time BLF, no simd
        start = timer()

        a.Assemble()
        end = timer()
        t_blf_normal += end-start

        ngsxfemglobals.SwitchSIMD(True)
        #########################
        # time linear form, simd

        start = timer()

        g.Assemble()
        end = timer()

        t_simd += end-start
        #########################
        # time BLF form, simd
        start = timer()

        b.Assemble()
        end = timer()
        t_blf_simd += end-start
    t_normal /= n
    t_blf_normal /= n
    t_simd /= n
    t_blf_simd /= n

    diffvec = f.vec.CreateVector()
    diffvec.data = f.vec - g.vec
    assert(diffvec.Norm() < 1e-8)

    diffvec = a.mat.AsVector().CreateVector()
    diffvec.data = a.mat.AsVector() - b.mat.AsVector()
    assert(diffvec.Norm() < 1e-8)

    # We cannot guarantee, that SIMD operations are faster then non-SIMD..
    #assert(t_normal > t_simd)
    print("Time for linear form, no simd: ", t_normal)
    print("Time for linear form, simd: ", t_simd)
    print("ratio: ", t_normal/t_simd)
    print("Time for blf, no simd: ", t_blf_normal)
    print("Time for blf, simd: ", t_blf_simd)
    print("ratio: ", t_blf_normal/t_blf_simd)


# space-time test case
@pytest.mark.parametrize("quad", [True, False])
@pytest.mark.parametrize("domain", [NEG, POS, IF])
def test_spacetime_integrateX_via_straight_cutted_quad2Dplus1D(domain, quad):
    print()
    print("spacetime test with quad=" + str(quad) + " and domain=" + str(domain))
    mesh = MakeStructured2DMesh(quads = quad, nx=1, ny=1)    

    tref = ReferenceTimeVariable()
    
    levelset = lambda t : 1 - 2*x - 2*t
    referencevals = { POS : 1./8, NEG : 1 - 1/8, IF : 1.0/2 }

    h1fes = H1(mesh,order=1)
    lset_approx_h1 = GridFunction(h1fes)
    tfe = ScalarTimeFE(1) 
    fes= SpaceTimeFESpace(h1fes,tfe)
    lset_approx = GridFunction(fes)

    InterpolateToP1(levelset(0),lset_approx_h1)
    lset_approx.vec[0:h1fes.ndof].data = lset_approx_h1.vec
    InterpolateToP1(levelset(1),lset_approx_h1)
    lset_approx.vec[h1fes.ndof:2*h1fes.ndof].data = lset_approx_h1.vec
    
    f = CoefficientFunction(1)
    n = 10
    t_ns = 0
    t_simd = 0
    error = 0
    for i in range(n):
        # non-SIMD
        ngsxfemglobals.SwitchSIMD(False)
        start = timer()
        integral_ns = IntegrateX(levelset_domain = { "levelset" : lset_approx, "domain_type" : domain, "time_order": 0, "order": 0},
                             mesh=mesh, cf=f)
        end = timer()
        t_ns += end - start

        # SIMD
        ngsxfemglobals.SwitchSIMD(True)
        start = timer()
        integral_simd = IntegrateX(levelset_domain = { "levelset" : lset_approx, "domain_type" : domain, "time_order": 0, "order": 0},
                             mesh=mesh, cf=f)
        end = timer()
        t_simd += end - start

        error += abs(integral_ns - integral_simd)
        #error += abs(integral - referencevals[domain])
    print("Non-SIMD to SIMD-ratio: ", t_ns/t_simd)
    assert error/n < 1e-8


# facetpatch test
@pytest.mark.parametrize("quad", [True, False])
def test_facetpatch(quad):
    print()
    print("test for FacetPatch with quad=" + str(quad))
    mesh = MakeStructured2DMesh(quads = quad, nx=2, ny=2)    

    fes = H1(mesh,order=3,dgjumps=True)
    u,v = fes.TnT()

    testruns = 5

    for i in range(testruns):
        a = BilinearForm(fes)
        a += u * v * dFacetPatch()
        # non-SIMD
        ngsxfemglobals.SwitchSIMD(False)
        start = timer()
        a.Assemble()
        end = timer()
        t_ns = end - start

        a2 = BilinearForm(fes)
        a2 += u * v * dFacetPatch()
        # SIMD
        ngsxfemglobals.SwitchSIMD(True)
        start = timer()
        a2.Assemble()
        end = timer()
        t_simd = end - start

    a.mat.AsVector().data -= a2.mat.AsVector().data

    error = Norm(a.mat.AsVector())
    #error += abs(integral - referencevals[domain])
    print("Non-SIMD to SIMD-ratio: ", t_ns/t_simd)
    assert error < 1e-9


# CalcLinearizedElementMatrix-test
def test_calc_linearized():
    print()
    print("Pytest for CalcLinearizedElementMatrix")
    square = SplineGeometry()
    square.AddRectangle([-1.5,-1.5],[1.5,1.5],bc=1)
    mesh = Mesh (square.GenerateMesh(maxh=0.8))
    
    levelset = (sqrt(sqrt(x*x*x*x+y*y*y*y)) - 1.0)

    lsetp1 = GridFunction(H1(mesh))
    lsetp1.Set(levelset)
    lset_neg = { "levelset" : lsetp1, "domain_type" : NEG, "subdivlvl" : 0}
    
    ci = CutInfo(mesh,lsetp1)
    hasneg = ci.GetElementsOfType(HASNEG)

    Vh = H1(mesh, order = 2, dirichlet=[])    
    active_dofs = GetDofsOfElements(Vh,hasneg)
    Vh = Compress(Vh,active_dofs)
    
    u,v = Vh.TnT()

    gfu = GridFunction(Vh)
    gfu.Set(sin(x))
    
    a1 = BilinearForm(Vh)
    a1 += SymbolicBFI(levelset_domain = lset_neg, form = 0.5*u**2 * v)
    a2 = BilinearForm(Vh)
    a2 += SymbolicBFI(levelset_domain = lset_neg, form = 0.5*u**2 * v)

    testruns = 5
    for i in range(testruns):
        # SIMD enabled
        ngsxfemglobals.SwitchSIMD(True)
        start = timer()
        a1.AssembleLinearization(gfu.vec)
        end = timer()

        t_simd = end - start

        # SIMD disabled
        ngsxfemglobals.SwitchSIMD(False)
        start = timer()
        a2.AssembleLinearization(gfu.vec)
        end = timer()
        
    t_ns = end - start

    print("Normal to SIMD ratio: ", str(t_ns/t_simd))

    w1 = gfu.vec.CreateVector()
    w2 = gfu.vec.CreateVector()

    w1.data = a1.mat * gfu.vec
    w2.data = a2.mat * gfu.vec
    
    # print(Norm(w1))
    # print(Norm(w2))

    w1.data -= w2.data
    diff = Norm(w1)
    print("diff : ",diff)
    assert diff < 1e-12


# test for ApplyElementMatrix
def test_calc_linearized():
    print()
    print("test for CalcLinearizedElementMatrix")
    square = SplineGeometry()
    square.AddRectangle([-1.5,-1.5],[1.5,1.5],bc=1)
    mesh = Mesh (square.GenerateMesh(maxh=0.8))
    
    levelset = (sqrt(sqrt(x*x*x*x+y*y*y*y)) - 1.0)

    lsetp1 = GridFunction(H1(mesh))
    lsetp1.Set(levelset)
    lset_neg = { "levelset" : lsetp1, "domain_type" : NEG, "subdivlvl" : 0}
    
    ci = CutInfo(mesh,lsetp1)
    hasneg = ci.GetElementsOfType(HASNEG)

    Vh = H1(mesh, order = 2, dirichlet=[])    
    active_dofs = GetDofsOfElements(Vh,hasneg)
    Vh = Compress(Vh,active_dofs)
    
    u,v = Vh.TnT()

    gfu = GridFunction(Vh)
    gfu.Set(sin(x))

    a1 = BilinearForm(Vh)
    a1 += SymbolicBFI(levelset_domain = lset_neg, form = 0.5*u**2 * v)
    a1.Assemble()
    
    a2 = BilinearForm(Vh)
    a2 += SymbolicBFI(levelset_domain = lset_neg, form = 0.5*u**2 * v)
    a2.Assemble()

    w1 = gfu.vec.CreateVector()
    w2 = gfu.vec.CreateVector()

    w1.data = a1.mat * gfu.vec
    w2.data = a2.mat * gfu.vec
    
    t_simd = 0
    t_ns = 0
    testruns = 5
    for i in range(testruns):
        # SIMD enabled
        ngsxfemglobals.SwitchSIMD(True)
        start = timer()
        a1.Apply(gfu.vec,w1)
        end = timer()
        t_simd += end-start

        # SIMD disabled
        ngsxfemglobals.SwitchSIMD(False)
        start = timer()
        a2.Apply(gfu.vec,w2)
        end = timer()
        t_ns += end-start

    print("Normal/SIMD: ", t_ns/t_simd)

    w1.data -= w2
    diff = Norm(w1)
    print("error: ",diff)
    assert diff < 1e-12


# copy from test_straightcutrule.py - this test compares the results SIMD to non-SIMD
@pytest.mark.parametrize("i", [2, 3, 4])
def test_eb_cut_integrator_2d(i):
    from netgen.geom2d import SplineGeometry

    ngsglobals.msg_level = 1

    levelset = (x -1) **4 + (y -1)**4 + 12*y + 1.4*(x -2)**3 - 15
    len_box = 4

    exact = sin(y)
    exact_grad = CoefficientFunction((0, cos(y)))

    order = 5

    l2errors = []
    h1errors = []

    extend_Fh = False
    condense = True

    square = SplineGeometry()
    square.AddRectangle([-len_box,-len_box],[len_box,len_box],bc=1)
    mesh = Mesh (square.GenerateMesh(maxh=len_box/1.5*0.5**(i), quad_dominated=False))
    print("Max_h = ", len_box/1.5*0.5**(i))

    lsetmeshadap = LevelSetMeshAdaptation(mesh, order=order, threshold=10.25, discontinuous_qn=True)
    deformation = lsetmeshadap.CalcDeformation(levelset)
    lsetp1 = lsetmeshadap.lset_p1
            
    # TraceFESpace 
    fes1 = L2(mesh, order=order)
    fes2 = FacetFESpace(mesh, order=order)
    fes3 = FacetFESpace(mesh, order=order - 1)
        
    Vhg = FESpace([fes1, fes2, fes3], dgjumps=not condense)
        
    # overwrite freedofs of VhG to mark only dofs that are involved in the cut problem
    ci = CutInfo(mesh, lsetp1)
    reg_Th = ci.GetElementsOfType(IF)   
    reg_Fh = GetFacetsWithNeighborTypes(mesh,a=reg_Th,b=reg_Th,use_and=not extend_Fh)

    gf_reg_Fh = GridFunction(FacetFESpace(mesh,order=0))
    for i in range(len(reg_Fh)):
        gf_reg_Fh.vec[i] = 1 if reg_Fh[i] else 0
        
    if not extend_Fh:
            if not condense:
                freedofs = CompoundBitArray( [GetDofsOfElements(fes1,reg_Th),  GetDofsOfFacets(fes2,reg_Fh) ,  GetDofsOfFacets(fes3,reg_Fh) ] )
            else:
                freedofs = CompoundBitArray( [fes1.FreeDofs(True) & GetDofsOfElements(fes1,reg_Th),  GetDofsOfFacets(fes2,reg_Fh) ,  GetDofsOfFacets(fes3,reg_Fh) ] )
    else:
            if not condense:
                freedofs = GetDofsOfElements(Vhg,reg_Th)
            else:
                freedofs = CompoundBitArray( [fes1.FreeDofs(True) & GetDofsOfElements(fes1,reg_Th),  GetDofsOfElements(fes2,reg_Th) ,  GetDofsOfElements(fes3,reg_Th) ] )
        
    normal_helper_gf = GridFunction(HDiv(mesh, order=0))
    for i in range(len(normal_helper_gf.vec)):
        normal_helper_gf.vec[i] = 1
    
    gfu = GridFunction(Vhg)
        
    #tangential projection to given normal
    def P(u,n_phi):
        return u - (u*n_phi)*n_phi
    
    #normalization (pointwise) of a vector
    def Normalized(u):
        return 1.0 / Norm(u) * u
    
    n_phi1 = Normalized(grad(lsetp1))
    n_phi2 = Normalized(grad(lsetp1).Other())
    
    h = specialcf.mesh_size
    n_F = specialcf.normal(2)
    
    conormal1 = Normalized(P(n_F,n_phi1))
    conormal2 = Normalized(P(-n_F,n_phi2))
    
    normal_helper_proj = InnerProduct(normal_helper_gf, n_F)*n_F
    normal_helper = Normalized(normal_helper_proj)
        
    def jump(u, uhat):
        return u - uhat
    
    beta_E = 4 * (order+1)**2
    beta_F = 100.
    beta_F2 = 1.
    lam_nd = 0.0 if order == 1 else 0.1*1./h + 0.1*h
    
    u, uhat, sigmahat = Vhg.TrialFunction()
    v, vhat, tauhat = Vhg.TestFunction()
    lset_if  = { "levelset" : lsetp1, "domain_type" : IF , "subdivlvl" : 0}
    
    a = RestrictedBilinearForm(Vhg,"a",reg_Th,reg_Fh,check_unused=False, flags={"eliminate_internal": condense})
    a += SymbolicBFI(levelset_domain = lset_if, form = P(grad(u),n_phi1) * P(grad(v),n_phi1) + u * v, definedonelements=reg_Th)
    a += SymbolicBFI(form = (lam_nd * grad(u)*n_phi1) * (grad(v)*n_phi1), definedonelements=reg_Th)
    
    a += SymbolicBFI(levelset_domain = lset_if, form = ( - InnerProduct(grad(u),conormal1)*jump(v, vhat)
                                                            - InnerProduct(grad(v),conormal1)*jump(u, uhat)
                                                            + beta_E/h * (u- uhat)*(v - vhat) ) *gf_reg_Fh,
                                                            element_boundary=True, definedonelements=reg_Th)
    
    a += SymbolicBFI(form = beta_F/(h*h) * (u- uhat) * (v - vhat) *gf_reg_Fh, element_boundary=True, definedonelements=reg_Th)

    a += SymbolicBFI(form = beta_F2 * (grad(u)*normal_helper- sigmahat) * (grad(v)*normal_helper - tauhat) *gf_reg_Fh, element_boundary=True, definedonelements=reg_Th)

    f_coeff = -(4*(y - 1)**3 + 12)**2*((4.2*(x - 2)**2 + 4*(x - 1)**3)**2 + (4*(y - 1)**3 + 12)**2)**(-1.0)*sin(y) + (4*(y - 1)**3 + 12)*((4.2*(x - 2)**2 + 4*(x - 1)**3)**2 + (4*(y - 1)**3 + 12)**2)**(-0.5)*(-12.0*(y - 1)**2*(4*(y - 1)**3 + 12)**2*((4.2*(x - 2)**2 + 4*(x - 1)**3)**2 + (4*(y - 1)**3 + 12)**2)**(-1.5) + 12*(y - 1)**2*((4.2*(x - 2)**2 + 4*(x - 1)**3)**2 + (4*(y - 1)**3 + 12)**2)**(-0.5) - 0.5*(4.2*(x - 2)**2 + 4*(x - 1)**3)**2*((4.2*(x - 2)**2 + 4*(x - 1)**3)**2 + (4*(y - 1)**3 + 12)**2)**(-1.5)*(16.8*x + 24*(x - 1)**2 - 33.6) + ((4.2*(x - 2)**2 + 4*(x - 1)**3)**2 + (4*(y - 1)**3 + 12)**2)**(-0.5)*(8.4*x + 12*(x - 1)**2 - 16.8))*cos(y) + 2*sin(y)
        
    f = LinearForm(Vhg)
    f += SymbolicLFI(levelset_domain = lset_if, form = f_coeff * v, definedonelements=reg_Th)

    mesh.SetDeformation(deformation)
    a.Assemble()
    f.Assemble();
        
    gfu.vec[:] = 0.0
    if not condense:
        gfu.vec.data = a.mat.Inverse(freedofs, "sparsecholesky") * f.vec
    else:
        f.vec.data += a.harmonic_extension_trans * f.vec 
        inv = a.mat.Inverse(freedofs, "sparsecholesky")
        gfu.vec.data = inv * f.vec
        
        gfu.vec.data += a.harmonic_extension * gfu.vec
        gfu.vec.data += a.inner_solve * f.vec
        
    err_sqr_coefs = (gfu.components[0]-exact)**2

    ngsxfemglobals.SwitchSIMD(True)
    l2error_simd = sqrt( Integrate( levelset_domain=lset_if, cf=err_sqr_coefs, mesh=mesh, order=2*order+1) )
        
    ngsxfemglobals.SwitchSIMD(False)
    l2error_ns = sqrt( Integrate( levelset_domain=lset_if, cf=err_sqr_coefs, mesh=mesh, order=2*order+1) )
        
    H1_err_sqr_coefs = Norm( P (gfu.components[0].Deriv() - exact_grad, n_phi1) )**2
    ngsxfemglobals.SwitchSIMD(True)
    h1error_simd = sqrt( l2error_simd**2 + Integrate( levelset_domain=lset_if, cf=H1_err_sqr_coefs, mesh=mesh, order=2*order+1) )
    ngsxfemglobals.SwitchSIMD(False)
    h1error_ns = sqrt( l2error_ns**2 + Integrate( levelset_domain=lset_if, cf=H1_err_sqr_coefs, mesh=mesh, order=2*order+1) )

    assert abs(l2error_simd - l2error_ns) < 1e-12
    assert abs(h1error_simd - h1error_ns) < 1e-12

    mesh.UnsetDeformation()

# copy from test_differential_symbol.pytest - this test compares the result SIMD to non-SIMD
@pytest.mark.parametrize('order', [1, 2, 3])
@pytest.mark.parametrize('DOM', [POS, NEG])
@pytest.mark.parametrize('skeleton', [True, False])
@pytest.mark.parametrize('element_boundary', [True, False])
def test_cut_symbols_dg(order, DOM, skeleton, element_boundary):
    mesh = Mesh(unit_square.GenerateMesh(maxh=0.2))
    levelset = (x - 0.5)**2 + (y - 0.5)**2 - 0.33**2

    if skeleton is False and element_boundary is False:
        return None
    if order > 1:
        lsetmeshadap = LevelSetMeshAdaptation(mesh, order=order,
                                              levelset=levelset)
    else:
        lsetmeshadap = NoDeformation(mesh, levelset)

    lset_h = lsetmeshadap.lset_p1
    deform = lsetmeshadap.deform

    ci = CutInfo(mesh, lset_h)
    els_hasdom = ci.GetElementsOfType(HAS(DOM))
    facets_dom = GetFacetsWithNeighborTypes(mesh, a=els_hasdom, b=els_hasdom)
    els_none = BitArray(mesh.ne)
    els_none[:] = False

    V = H1(mesh, order=order, dgjumps=True)
    u, v = V.TnT()

    gfu = GridFunction(V)
    gfu.Set(sin(x))

    w1, w2 = gfu.vec.CreateVector(), gfu.vec.CreateVector()

    # Bilinear form to integrate
    nF = specialcf.normal(mesh.dim)
    flux_u = -0.5 * (grad(u) + grad(u.Other())) * nF
    flux_v = -0.5 * (grad(v) + grad(v.Other())) * nF
    jump_u = u - u.Other()
    jump_v = v - v.Other()

    form = jump_u * jump_v + flux_u * jump_v + flux_v * jump_u

    # Differential Symbol Version
    a1 = RestrictedBilinearForm(V, element_restriction=els_hasdom,
                                facet_restriction=facets_dom,
                                check_unused=False)
    a1 += form * dCut(lset_h, DOM, definedonelements=facets_dom,
                      deformation=deform, skeleton=skeleton,
                      element_boundary=element_boundary)
    a2 = RestrictedBilinearForm(V, element_restriction=els_hasdom,
                                facet_restriction=facets_dom,
                                check_unused=False)
    a2 += form * dCut(lset_h, DOM, definedonelements=facets_dom,
                      deformation=deform, skeleton=skeleton,
                      element_boundary=element_boundary)
    
    ngsxfemglobals.SwitchSIMD(True)
    a1.Assemble()
    ngsxfemglobals.SwitchSIMD(False)
    a2.Assemble()
    
    w1.data = a1.mat * gfu.vec
    w2.data = a2.mat * gfu.vec

    # # SymbolicBFI version
    # lset_dom = {'levelset': lset_h, 'domain_type': DOM, 'subdivlvl': 0}
    # a2 = RestrictedBilinearForm(V, element_restriction=els_hasdom,
    #                             facet_restriction=facets_dom,
    #                             check_unused=False)
    # a2 += SymbolicBFI(levelset_domain=lset_dom, form=form,
    #                   definedonelements=facets_dom, skeleton=skeleton,
    #                   element_boundary=element_boundary)
    # mesh.SetDeformation(deform)
    # a2.Assemble()
    # mesh.UnsetDeformation()
    # w2.data = a2.mat * gfu.vec

    # Check
    w1.data -= w2
    diff = Norm(w1)
    print(f'diff : {diff}')
    assert diff < 1e-12


"""
	Description:
	------------
	Sharp cube defined by six level sets, one for each face.
"""

# Import libraries
from netgen.csg import *
from ngsolve import *
from xfem import *
from xfem.mlset import *

# Mesh
geo = CSGeometry()
geo.Add(OrthoBrick(Pnt(-0.8,-0.8,-0.8), Pnt(0.8,0.8,0.8)))
mesh = Mesh(geo.GenerateMesh(maxh=1))

# Level sets
V = H1(mesh,order=1)
lsetp1_x_upper, lsetp1_x_lower = GridFunction(V), GridFunction(V)
lsetp1_y_upper, lsetp1_y_lower = GridFunction(V), GridFunction(V)
lsetp1_z_upper, lsetp1_z_lower = GridFunction(V), GridFunction(V)

InterpolateToP1( x - 0.5, lsetp1_x_upper)
InterpolateToP1( x + 0.5, lsetp1_x_lower)
InterpolateToP1( y - 0.5, lsetp1_y_upper)
InterpolateToP1( y + 0.5, lsetp1_y_lower)
InterpolateToP1( z - 0.5, lsetp1_z_upper)
InterpolateToP1( z + 0.5, lsetp1_z_lower)

# Integrate
cube = DomainTypeArray(dtlist=[(NEG, POS, NEG, POS, NEG, POS)])

# Compute volume
volume = Integrate(levelset_domain={"levelset": [lsetp1_x_upper, lsetp1_x_lower, 
                                                 lsetp1_y_upper, lsetp1_y_lower, 
                                                 lsetp1_z_upper, lsetp1_z_lower],
                                    "domain_type": cube.as_list},
                              mesh=mesh, cf=1, order=0)

print("volume = {:10.8f}".format(volume))            
print("volume error = {:4.3e}".format(abs(volume - 1)))
assert abs(volume - 1)


# Compute are of surfaces
for i in range(6):
    domian = [NEG, POS, NEG, POS, NEG, POS]
    domian[i] = IF

    area = Integrate(levelset_domain = {"levelset" : [lsetp1_x_upper, lsetp1_x_lower, 
                                                      lsetp1_y_upper, lsetp1_y_lower, 
                                                      lsetp1_z_upper, lsetp1_z_lower],
                                        "domain_type" : tuple(domian)},
                     mesh=mesh, cf=1, order=0)

    print("area_face{:d} = {:10.8f}".format(i, area))                     
    print("area_face{:d} error = {:4.2e}".format(i, abs(area - 1)))
    assert abs(area-1) < 1e-12

surface_area = Integrate(levelset_domain={"levelset": [lsetp1_x_upper, lsetp1_x_lower, 
                                                 lsetp1_y_upper, lsetp1_y_lower, 
                                                 lsetp1_z_upper, lsetp1_z_lower],
                                    "domain_type": cube.Boundary().as_list},
                              mesh=mesh, cf=1, order=0)

print("surface_area = {:10.8f}".format(surface_area))            
print("surface_area error = {:4.3e}".format(abs(surface_area - 6)))
assert abs(surface_area - 6)


# Compute length of lines
sides = [(IF, POS, IF, POS, NEG, POS), (IF, POS, NEG, IF, NEG, POS),
         (IF, POS, NEG, POS, IF, POS), (IF, POS, NEG, POS, NEG, IF),
         (NEG, IF, IF, POS, NEG, POS), (NEG, IF, NEG, IF, NEG, POS),
         (NEG, IF, NEG, POS, IF, POS), (NEG, IF, NEG, POS, NEG, IF),
         (NEG, POS, IF, POS, IF, POS), (NEG, POS, NEG, IF, IF, POS),
         (NEG, POS, IF, POS, NEG, IF), (NEG, POS, NEG, IF, NEG, IF)]

for i, side in enumerate(sides):
    length = Integrate(levelset_domain = {"levelset": [lsetp1_x_upper, lsetp1_x_lower, 
                                                      lsetp1_y_upper, lsetp1_y_lower, 
                                                      lsetp1_z_upper, lsetp1_z_lower],
                                        "domain_type": side},
                     mesh=mesh, cf=1, order=0)

    
    print("length{:d} = {:10.8f}".format(i, length))                     
    print("length{:d} error = {:4.2e}".format(i, abs(length - 1)))
    assert abs(length - 1) < 1e-12


# Point evaluation
point_domains = {"ppp": (IF, POS, IF, POS, IF, POS),
                 "ppn": (IF, POS, IF, POS, NEG, IF),
                 "pnp": (IF, POS, NEG, IF, IF, POS), 
                 "npp": (NEG, IF, IF, POS, IF, POS),
                 "pnn": (IF, POS, NEG, IF, NEG, IF),
                 "npn": (NEG, IF, IF, POS, NEG, IF),
                 "nnp": (NEG, IF, NEG, IF, IF, POS),
                 "nnn": (NEG, IF, NEG, IF, NEG, IF) }

vals = {"ppp": 1.5, "ppn": 0.5, "pnp": 0.5, "npp": 0.5, "pnn": -0.5, 
        "npn": -0.5, "nnp": -0.5, "nnn": -1.5 }
    
for i, (key, domain) in enumerate(point_domains.items()):

    point = Integrate(levelset_domain = {"levelset": [lsetp1_x_upper, lsetp1_x_lower, 
                                                      lsetp1_y_upper, lsetp1_y_lower, 
                                                      lsetp1_z_upper, lsetp1_z_lower],
                                        "domain_type": domain},
                     mesh=mesh, cf=x+y+z, order=0)

    
    print("point{:d} = {:10.8f}".format(i, point))                     
    print("point{:d} error = {:4.2e}".format(i, abs(point-vals[key])))
    assert abs(point - vals[key]) < 1e-12



"""
In this example we solve a scalar *unfitted* interface problem. As a
discretisation method we use a level set based geometry description and
a Cut or Extended Finite element method with a Nitsche formulation to
impose the interface conditions.

Domain:
-------
The domain is [-1.5,1.5]^2 while the interface is described by a level
set function ( unit ball in the ||·||_4 norm ). In the discretisation
the level set function is approximated with a piecewise linear
interpolation.

PDE problem:
------------
domain equations:
    - alpha_i (u_xx + u_yy) =   f in sub-domain i, i=1,2
interface conditions:
                        [u] =    0 on interface (continuity across
                                                 the interface     ),
           [-alpha · du/dn] =    0 on interface (conservation of the
                                                 (diffusive) flux. ),
                         u  =  u_D on domain boundary.

The r.h.s. term f and the Dirichlet data u_D is chosen according to a
manufactured solution which allows us to measure errors after the
computation of a discrete solution. The coefficients alpha are
domain-wise constants which are different in the two sub-domains.

Discretisation:
---------------
* Finite element space: We consider two different formulations:
  * "CutFEM" For each sub-domain we consider the space of piecewise
    polynomials restricted to the corresponding sub-domain.
  * "XFEM" we consider a standard space of piecewise polynomials combined
    with an enrichment space.

* Variational formulation: We use a Nitsche formulation which involves
  averages of the fluxes and jumps of the solution across the interface.
  For the average we use the Hansbo-choice [1] where the average is
  adjusted to the local cut configuration in order to ensure stability
  of the resulting formulation.

Implementational aspects:
-------------------------
* Geometry approximation: To approximate the implicitly described
  geometry we use the piecewise (multi-) linear interpolation of the
  level set function. For this geometry approximation (arbitrary order)
  accurate numerical integration is provided.
  For higher order approximation an additional mesh deformation is
  applied that maps the low order interface approximation to something
  more accurate. See also jupyter tutorials `intlset` and `cutfem`

* Linear systems: A (sparse) direct solver is applied to solve the
  arising linear systems.

Literature:
-----------
[1] A. Hansbo, P. Hansbo, An unfitted finite element method, based on
    Nitsche's method, for elliptic interface problems, Comp. Meth.
    Appl. Mech. Eng., 191(47):5537-5552, 2002.
"""

# ------------------------------ LOAD LIBRARIES -------------------------------
from netgen.geom2d import SplineGeometry
from ngsolve import *
from xfem import *
from xfem.lsetcurv import LevelSetMeshAdaptation

from math import pi

# -------------------------------- PARAMETERS ---------------------------------
# Domain corners
ll, ur = (-1.5, -1.5), (1.5, 1.5)
# Mesh size
maxh = 0.2
# Finite element space order
order = 2

# Diffusion coefficients for the sub-domains (NEG/POS):
alpha = [1.0, 2.0]
# Nitsche penalty parameter
lambda_nitsche = 20

formulation = "XFEM"  # or "CUTFEM"

# ----------------------------------- MAIN ------------------------------------
# We generate the background mesh of the domain and use a simplicity
# triangulation to obtain a mesh with quadrilaterals use
# 'quad_dominated=True'
square = SplineGeometry()
square.AddRectangle(ll, ur, bc=1)
mesh = Mesh(square.GenerateMesh(maxh=maxh, quad_dominated=False))

# Manufactured solution and corresponding r.h.s. data CoefficientFunctions:
r22 = x**2 + y**2
r44 = x**4 + y**4
r66 = x**6 + y**6
r41 = sqrt(sqrt(r44))
r4m3 = 1.0 / r41**3
solution = [1 + pi / 2 - sqrt(2.0) * cos(pi / 4 * r44), pi / 2 * r41]
coef_f = [-alpha[i] * (solution[i].Diff(x).Diff(x)
                       + solution[i].Diff(y).Diff(y)) for i in range(2)]

# Level set function of the domain (phi = ||x||_4 - 1) and its interpolation:
levelset = r41 - 1.0

if order > 1:
    lsetadap = LevelSetMeshAdaptation(mesh, order=order, levelset=levelset)
else:
    lsetadap = NoDeformation(mesh, levelset)
lsetp1 = lsetadap.lset_p1

# Background FESpaces (used as CutFESpaces later-on):
Vh = H1(mesh, order=order, dirichlet=".*")

# Gathering information on cut elements:
ci = CutInfo(mesh, lsetp1)

if formulation == "XFEM":
    Vhx = XFESpace(Vh, lsetp1)
    VhG = Vh * Vhx

    gfu = GridFunction(VhG)

    (u_std, u_x), (v_std, v_x) = VhG.TnT()

    u = [u_std + op(u_x) for op in [neg, pos]]
    v = [v_std + op(v_x) for op in [neg, pos]]
    gradu = [grad(u_std) + op(u_x) for op in [neg_grad, pos_grad]]
    gradv = [grad(v_std) + op(v_x) for op in [neg_grad, pos_grad]]

    gfu_std, gfu_x = gfu.components
    gfuh = [gfu_std + op(gfu_x) for op in [neg, pos]]
else:
    # Create CutFESpace by selecting only active elements for inside
    # and outside:
    VhG = Compress(Vh, GetDofsOfElements(Vh, ci.GetElementsOfType(HASNEG))) \
        * Compress(Vh, GetDofsOfElements(Vh, ci.GetElementsOfType(HASPOS)))

    gfu = GridFunction(VhG)

    u, v = VhG.TnT()
    gradu, gradv = [[grad(w[i]) for i in [0, 1]] for w in [u, v]]

    gfuh = gfu.components

print("unknowns in background FESpace : ", Vh.ndof)
print("unknowns in " + formulation + " FESpace : ", VhG.ndof)


# Coefficients / parameters:
n = 1.0 / grad(lsetp1).Norm() * grad(lsetp1)
h = specialcf.mesh_size
# The cut ratio extracted from the cutinfo-class
kappa = (CutRatioGF(ci), 1.0 - CutRatioGF(ci))
# Nitsche stabilization parameter:
stab = lambda_nitsche * (alpha[1] + alpha[0]) / h

average_flux_u = sum([- kappa[i] * alpha[i] * gradu[i] * n for i in [0, 1]])
average_flux_v = sum([- kappa[i] * alpha[i] * gradv[i] * n for i in [0, 1]])

# Integration domains for integration on negative/positive sub-domains
# and on the interface: Here, the integration is (geometrically) exact
# if the "levelset"-argument is a piecewise (multi-)linear function.
# We further provide a mesh deformation that is applied in the higher order
# case:
dx = tuple([dCut(lsetp1, dt, deformation=lsetadap.deform,
                 definedonelements=ci.GetElementsOfType(HAS(dt)))
            for dt in [NEG, POS]])
ds = dCut(lsetp1, IF, deformation=lsetadap.deform)

# Bilinear form for the unfitted Nitsche formulation:
a = BilinearForm(VhG, symmetric=True)
a += sum(alpha[i] * gradu[i] * gradv[i] * dx[i] for i in [0, 1])
a += (average_flux_u * (v[0] - v[1]) + average_flux_v * (u[0] - u[1])
      + stab * (u[0] - u[1]) * (v[0] - v[1])) * ds

# R.h.s.:
f = LinearForm(VhG)
f += sum(coef_f[i] * v[i] * dx[i] for i in [0, 1])

# Setting domain boundary conditions:
# The context manager lsetadap applies the mesh deformation
# in the higher order case:
with lsetadap:
    if formulation == "XFEM":
        gfu.components[0].Set(solution[1], BND)
    else:
        gfu.components[1].Set(solution[1], BND)

# Setting up matrix and vector
a.Assemble()
f.Assemble()

# Homogenization of boundary data and solution of linear system
f.vec.data -= a.mat * gfu.vec
gfu.vec.data += a.mat.Inverse(VhG.FreeDofs()) * f.vec

# Visualize levelset, interpolated levelset and discrete solution:
Draw(levelset, mesh, "levelset")
Draw(lsetp1, mesh, "levelset_P1")
# Note that standard netgen-gui visualization does not respect
# the discontinuities, they are smeared out. To see kinks or jumps
# more clearly increase the subdivision option in the visualization.
if formulation == "CUTFEM":
    DrawDC(lsetp1, gfu.components[0], gfu.components[1],
           mesh, "u", deformation=lsetadap.deform)
    Draw(IfPos(-lsetp1,
               alpha[0]*grad(gfu.components[0]),
               alpha[1]*grad(gfu.components[1])),
         mesh, "sigma", deformation=lsetadap.deform)
else:
    DrawDC(lsetp1,
           gfu.components[0]+neg(gfu.components[1]),
           gfu.components[0]+pos(gfu.components[1]),
           mesh, "u", deformation=lsetadap.deform)
    Draw(IfPos(-lsetp1,
               alpha[0]*grad(gfu.components[0])+neg_grad(gfu.components[1]),
               alpha[1]*grad(gfu.components[0])+pos_grad(gfu.components[1])),
         mesh, "sigma", deformation=lsetadap.deform)

# Computation of L2 error:
err_sqr = sum([(gfuh[i] - solution[i])**2 * dx[i].order(2 * order)
               for i in [0, 1]])
print("L2 error : ", sqrt(Integrate(err_sqr, mesh)))



import pytest
from ngsolve import *
from xfem import *
from ngsolve.meshes import *
from netgen.geom2d import SplineGeometry
from netgen.csg import CSGeometry, OrthoBrick, Pnt
from math import pi
from xfem.lsetcurv import *

def test_apply():
    square = SplineGeometry()
    square.AddRectangle([-1.5,-1.5],[1.5,1.5],bc=1)
    mesh = Mesh (square.GenerateMesh(maxh=0.8))
    
    levelset = (sqrt(sqrt(x*x*x*x+y*y*y*y)) - 1.0)

    lsetp1 = GridFunction(H1(mesh))
    lsetp1.Set(levelset)
    lset_neg = { "levelset" : lsetp1, "domain_type" : NEG, "subdivlvl" : 0}
    
    ci = CutInfo(mesh,lsetp1)
    hasneg = ci.GetElementsOfType(HASNEG)

    Vh = H1(mesh, order = 2, dirichlet=[])    
    active_dofs = GetDofsOfElements(Vh,hasneg)
    Vh = Compress(Vh,active_dofs)
    
    u,v = Vh.TnT()
    
    a = BilinearForm(Vh)
    a += SymbolicBFI(levelset_domain = lset_neg, form = u * v)
    a += SymbolicBFI(levelset_domain = lset_neg, form = grad(u)[0] * v)
    
    gfu = GridFunction(Vh)
    gfu.Set(1)

    a.Assemble()

    Au1 = gfu.vec.CreateVector()
    Au2 = gfu.vec.CreateVector()

    Au1.data = a.mat * gfu.vec
    
    a.Apply(gfu.vec,Au2)
    
    Au1.data -= Au2
    diff = Norm(Au1)
    print("diff : ",diff)
    assert diff < 1e-12

    
def test_nonlin():
    square = SplineGeometry()
    square.AddRectangle([-1.5,-1.5],[1.5,1.5],bc=1)
    mesh = Mesh (square.GenerateMesh(maxh=0.2, quad_dominated=False))

    levelset = (sqrt(sqrt(x*x*x*x+y*y*y*y)) - 1.0)
    lsetp1 = GridFunction(H1(mesh,order=1))
    InterpolateToP1(levelset,lsetp1)

    Vh = H1(mesh, order=1, dirichlet=[1,2,3,4], dgjumps=True)

    ci = CutInfo(mesh, lsetp1)

    hasneg = ci.GetElementsOfType(HASNEG)  # <- "hasneg": has (also) negative level set values
    hasif = ci.GetElementsOfType(IF)
    
    ba_facets = GetFacetsWithNeighborTypes(mesh,a=hasneg,b=hasif)

    freedofs = Vh.FreeDofs()
    # freedofs &= CompoundBitArray([GetDofsOfElements(Vh,hasneg),GetDofsOfElements(Vh,haspos)])
    freedofs &= GetDofsOfElements(Vh,hasneg)

    u = Vh.TrialFunction()
    v = Vh.TestFunction()

    lset_neg = { "levelset" : lsetp1, "domain_type" : NEG, "subdivlvl" : 0}

    a = BilinearForm(Vh, symmetric = True)
    a += SymbolicBFI(levelset_domain = lset_neg, form = (u+u**2-1)*v)
    a += SymbolicFacetPatchBFI(form = (u-u.Other())*(v-v.Other()),
                               skeleton=False,
                               definedonelements=ba_facets)

    gfu = GridFunction(Vh)

    from ngsolve.solvers import Newton

    Newton(a,gfu,freedofs=freedofs, inverse="")

    
if __name__ == "__main__":
    test_apply()
    test_nonlin()



# ------------------------------ LOAD LIBRARIES -------------------------------
import pandas as pd
from ngsolve import *
from netgen.geom2d import SplineGeometry
from xfem import *
from math import pi
from xfem.lset_spacetime import *
ngsglobals.msg_level = 1


# -------------------------------- PARAMETERS ---------------------------------
# DISCRETIZATION PARAMETERS:
def CalcGeomError(k_t, k_s, i, j):
    n_steps = 2**i
    space_refs = j
    # Polynomial order in time for level set approximation
    lset_order_time = k_t
    # Integration order in time
    time_order = 2 * k_t
    # Time stepping parameters
    tstart = 0
    tend = 0.5
    delta_t = (tend - tstart) / n_steps
    maxh = 0.5
    # Ghost-penalty parameter
    gamma = 0.05
    # Map from reference time to physical time
    told = Parameter(tstart)
    t = told + delta_t * tref

    # PROBLEM SETUP:

    # Outer domain:
    rect = SplineGeometry()
    rect.AddRectangle([-0.6, -1], [0.6, 1])

    # Level set geometry
    # Radius of disk (the geometry)
    R = 0.5
    # Position shift of the geometry in time
    rho = (1 / (pi)) * sin(2 * pi * t)
    # Convection velocity:
    w = CoefficientFunction((0, rho.Diff(t)))
    # Level set
    r = sqrt(x**2 + (y - rho)**2)
    levelset = r - R

    # ------------------------------- MAIN ------------------------------------
    ngmesh = rect.GenerateMesh(maxh=maxh, quad_dominated=False)
    for j in range(space_refs):
        ngmesh.Refine()
    mesh = Mesh(ngmesh)
    # Space time version of Levelset Mesh Adapation object. Also offers
    # integrator helper functions that involve the correct mesh deformation
    lsetadap = LevelSetMeshAdaptation_Spacetime(mesh, order_space=k_s,
                                                order_time=lset_order_time,
                                                threshold=0.5,
                                                lset_lower_bound=-3*delta_t,
                                                lset_upper_bound=3*delta_t,
                                                discontinuous_qn=True)

    def dt(u):
        return 1.0 / delta_t * dtref(u)

    maxdist = 0

    while tend - told.Get() > delta_t / 2:
        lsetadap.CalcDeformation(levelset)
        mesh.SetDeformation(lsetadap.deform)
        maxdist = max(lsetadap.CalcMaxDistance(levelset), maxdist)
        mesh.UnsetDeformation()
        told.Set(told.Get()+delta_t)
    print(maxdist)
    return maxdist


k_t = 3
k_s = 3
df = pd.DataFrame()

for i in range(6):
    for j in range(i, i+1):
        maxdist = CalcGeomError(k_t, k_s, i, j)
        df = df.append({'i': i, 'j': j, 'maxdist': maxdist}, ignore_index=True)
df['i'] = df['i'].astype(int)
df['j'] = df['j'].astype(int)
df.to_csv("geomerror.csv", index=False)
print(df)



"""
example of a square described by two level set functions living completely on one element
"""

# import netgen and xfem stuff
from netgen.geom2d import SplineGeometry
from ngsolve import *
from xfem import *


# Generate mesh
square = SplineGeometry()
square.AddRectangle([0,0], [1,1], bc=1)
mesh = Mesh (square.GenerateMesh(maxh=2, quad_dominated=False))

# Levelsets
lsetp1_a = GridFunction(H1(mesh,order=1))
lsetp1_b = GridFunction(H1(mesh,order=1))

InterpolateToP1( x - 0.5,lsetp1_a)
InterpolateToP1( y - 0.5,lsetp1_b)

Draw (lsetp1_a, mesh, "lset_a")
Draw (lsetp1_b, mesh, "lset_b")
Draw (lsetp1_a * lsetp1_b, mesh, "lset_mult")


# First test: A little square on the right bottom
area = Integrate(levelset_domain={"levelset": [lsetp1_a,lsetp1_b], 
                                  "domain_type": (POS,NEG)},
                 mesh=mesh, cf=1, order=0)

print("area = {:10.8f}".format(area))                     
print("area error = {:4.3e}".format(abs(area-0.25)))     


length1 = Integrate(levelset_domain={"levelset": [lsetp1_a, lsetp1_b], 
                                     "domain_type": (POS,IF)},
                    mesh=mesh, cf=1, order=0)

print("length1 =", length1)                     
print("length1 error =", abs(length1-0.5))
assert abs(length1-0.5) < 1e-12

length2 = Integrate(levelset_domain={"levelset": [lsetp1_a, lsetp1_b], 
                                     "domain_type": [(POS, IF), (IF, NEG)]},
                    mesh=mesh, cf=1, order=0)

print("length2 =", length2)                     
print("length2 error =", abs(length2-1))
assert abs(length2-1) < 1e-12


# solve the Poisson equation -Delta u = f
# with Dirichlet boundary condition u = 0

from ngsolve import *
from netgen.geom2d import unit_square
from netgen.geom2d import SplineGeometry
from xfem import *

ngsglobals.msg_level = 1

# generate a triangular mesh of mesh-size 0.2
square = SplineGeometry()
square.AddRectangle([-1.5,-1.5],[1.5,1.5],bc=1)
mesh = Mesh (square.GenerateMesh(maxh=0.2, quad_dominated=False))

mu = [1, 1e-7]
order = 1


rhsscal = -4*mu[0]*mu[1] 
coef_f = [ rhsscal, rhsscal ]
distsq = x*x + y*y - 1
solution = [ mu[1]* distsq, mu[0] * distsq ]

levelset = ( sqrt(x*x + y*y) - 1 )
subdivlvl = 0

def main():
    print('hallo')
    gamma_stab = 10
    nref = 2

    params =    {  
                    "order"     : order,
                    "mu"        : mu,
                    "gamma_stab": gamma_stab,
                    "kappa"     : "harmonic" # hansbo, highorder
                }


    # levelset fe space
    VhL = H1(mesh, order = order, dirichlet=[], dgjumps = True)
    lsetp1 = GridFunction(VhL)
    InterpolateToP1(levelset,lsetp1)
    ci = CutInfo(mesh, lsetp1)   
    

    Vh = H1(mesh, order=order, dirichlet=[1,2,3,4])
    VhNeg = Compress( Vh, active_dofs = GetActiveDof(mesh, HASNEG) )
    VhPos = Compress( Vh, active_dofs = GetActiveDof(mesh, HASPOS) )

    CutVh = FESpace( [VhNeg, VhPos], flags={"dgjumps": True} )

    # diffplot = GridFunction(CutVh)

    # ifdofs = IfaceUnks(CutVh,ci)

    # for i in range(len(diffplot.vec)):
    #     if ( ifdofs[i] ):
    #         diffplot.vec[i] = 1
    # ucoef = IfPos( lsetp1, diffplot.components[1], diffplot.components[0] )
    # Draw(ucoef, mesh, "diff")
    # print(ifdofs)
    # ifdofs &= CutVh.FreeDofs()
    # print(ifdofs)


    # return

    prolongNeg = P1Prolongation(mesh)
    prolongPos = P1Prolongation(mesh)
    CutProl = CompoundProlongation( CutVh )
    CutProl.AddProlongation(prolongNeg)
    CutProl.AddProlongation(prolongPos)
    prolongNeg.Update(VhNeg)
    prolongPos.Update(VhPos)
    

    a,f = AssembleCutPoisson( CutVh=CutVh, ci=ci, lsetp1=lsetp1, params=params )

    preJ = a.mat.CreateSmoother(CutVh.FreeDofs())
    GSpre = GaussSeid(preJ)

    #usol = solvers.PreconditionedRichardson(a,f.vec,pre=GSpre,maxit=500,tol=1e-6)

    #list of coarse grid matrices
    mats = {}
    mats[0] = a.mat
    #list of smoothers
    smoothers = {}
    smoothers[0] = CutFemSmoother(a, CutVh, ci) #preJ#GSpre
    #coarse grid solver
    inv = a.mat.Inverse(CutVh.FreeDofs(), inverse="sparsecholesky")
    
    for i in range(nref):
        mesh.Refine()
        #levelset update
        VhL.Update()
        lsetp1.Update()
        InterpolateToP1(levelset,lsetp1)
        ci = CutInfo(mesh,lsetp1)
        #cut space update
        Vh.Update()
        CutVh.components[0].SetActiveDofs( GetActiveDof(mesh,HASNEG) )
        CutVh.components[1].SetActiveDofs( GetActiveDof(mesh,HASPOS) )
        CutVh.Update()
        CutProl.Update(CutVh)
        #reassemble
        a,f = AssembleCutPoisson( CutVh=CutVh, ci=ci, lsetp1=lsetp1, params=params )        
        mats[i+1] = a.mat

        #preJ = a.mat.CreateSmoother(CutVh.FreeDofs())
        #GSpre = GaussSeid(preJ)
        smoothers[i+1] = CutFemSmoother(a, CutVh, ci) #preJ#GSpre        

    errors = [f.vec.CreateVector() for i in range(nref+1)]
    defects = [f.vec.CreateVector() for i in range(nref+1)]

    def MultiGridIter(rhs,nu,u,level):
        error = errors[level]
        defect = defects[level]

        if level == 0 :
            u.data = inv*rhs
            return
    #smoothing
        for i in range(nu):
            smoothers[level].Smooth(u,rhs)
    #compute defect
        defect.data = rhs - mats[level]*u
    #restrict defect
        CutProl.Restrict(level,defect)

        error[:] = 0
        MultiGridIter(defect,nu,error,level-1)

        CutProl.Prolongate(level,error)

        u.data += error
    #smoothing
        for i in range(nu):
            smoothers[level].Smooth(u,rhs)
        return u

    class MultiGridCL(BaseMatrix):
        def __init__(self,nu,l):
            super(MultiGridCL,self).__init__()
            self.nu = nu
            self.l = l
        def Mult(self,w,z):
            #z[:] = 0.0
            #z.data = w
            #z.data = 
            MultiGridIter(w,self.nu,z,self.l)
        def Height(self):
            return a.mat.height
        def Width(self):
            return a.mat.height

    MGpre = MultiGridCL(2,nref)

    usol = GridFunction(CutVh) 
    usol.vec[:] = 0.
    #set boundary condiitions
    #usol.components[1].Set(solution[1], BND)

    #usol.vec.data = mats[0] * usol.vec


    res = usol.vec.CreateVector()
    projres = usol.vec.CreateVector()
    tol = 1e-6
    proj = Projector(mask=CutVh.FreeDofs(),range=True) #projectors[nref]
    fnew = f.vec.data     

    projres.data = proj*fnew #f.vec
    normf = Norm(projres)

    for it in range(1,20):
        usol.vec.data = MGpre*fnew
        res.data = fnew - a.mat*usol.vec
        projres.data = proj*res
        res_norm = Norm(projres) / normf
        print("it =", it, " ||res||_2 =", res_norm)
        if res_norm < tol:
            break

    udraw = GridFunction( CutVh )
    udraw.components[1].Set( solution[1], BND )
    udraw.vec.data += usol.vec

    # Computation of L2 error:
    err_sqr_coefs = [(udraw.components[i]-solution[i])*(udraw.components[i]-solution[i]) for i in [0,1] ]
    lset_doms = LsetDoms( levelset, lsetp1, subdivlvl )
    l2error = sqrt( sum( [Integrate(lset_doms[i], cf=err_sqr_coefs[i], mesh=mesh, order=2*order, heapsize=1000000) for i in [0,1] ]))   
    print("L2 error : ",l2error)

    # pretty printing    
    ucoef = IfPos( lsetp1, udraw.components[1], udraw.components[0] )    
    #ucoef = IfPos( lsetp1, solution[1], solution[0] )    
    Draw (ucoef,mesh,'u')
    

class GaussSeid(BaseMatrix):
    def __init__ (self, smoother):
        super(GaussSeid, self).__init__()
        self.smoother = smoother
    def Mult (self, x, y):
        self.smoother.Smooth(y, x)
        #self.smoother.SmoothBack(y,x)
    def Height (self):
        return self.smoother.height
    def Width (self):
        return self.smoother.height

def GetActiveDof(mesh,HASPOSNEG):
        Vh = H1(mesh, order = order, dirichlet=[], dgjumps = True)
        lsetp1 = GridFunction(Vh)
        InterpolateToP1(levelset,lsetp1)
        ci = CutInfo(mesh,lsetp1)
        return GetDofsOfElements(Vh,ci.GetElementsOfType(HASPOSNEG))


def CutFESpace( V, ci, flags=None ):    
    hasneg = ci.GetElementsOfType(HASNEG)  
    haspos = ci.GetElementsOfType(HASPOS)

    if isinstance(V,list):
        if len(V) != 2:
            raise ValueError("you need to provide a list of two FE spaces")
        else:
            return FESpace( [   Compress(V[0],active_dofs=GetDofsOfElements(V[0],hasneg) ), 
                                Compress(V[1],active_dofs=GetDofsOfElements(V[1],haspos) ) 
                            ], flags=flags
                          ) 
    else:
        return FESpace( [   Compress(V,active_dofs=GetDofsOfElements(V,hasneg) ), 
                            Compress(V,active_dofs=GetDofsOfElements(V,haspos) ) 
                        ], flags=flags
                      ) 

def LsetDoms( levelset, lsetp1, subdivlvl ):
    if (subdivlvl == 0):
        lsetint = lsetp1
    else:
        lsetint = levelset

    lset_neg = { "levelset" : lsetint, "domain_type" : NEG, "subdivlvl" : subdivlvl}
    lset_pos = { "levelset" : lsetint, "domain_type" : POS, "subdivlvl" : subdivlvl}
    lset_if  = { "levelset" : lsetint, "domain_type" : IF , "subdivlvl" : subdivlvl}

    lset_dom = [ lset_neg, lset_pos,  lset_if ]

    return lset_dom

def KappaLambda(params,ci):
    kappac  = params["kappa"]
    mu      = params["mu"]
    order   = params["order"]

    if kappac == "harmonic":
        kappa = [ mu[1] / ( mu[0] + mu[1] ) , mu[0] /  ( mu[0] + mu[1] ) ]
        # stabilization parameter for Nitsche
        lambda_nitsche  = 2*mu[0]*mu[1]/(mu[0] + mu[1]) * 10 * order * order
    elif kappac == "hansbo":
        kappa = [CutRatioGF(ci),1.0-CutRatioGF(ci)]
        # stabilization parameter for Nitsche
        lambda_nitsche  = 0.5 * (mu[0]+mu[1]) * 20 * (order+1) * order
    elif kappac == "highorder":
        kappaHO = IfPos( CutRatioGF(ci) -0.5, 1, 0 )
        kappa = [kappaHO,1.0-kappaHO]
        # stabilization parameter for Nitsche
        lambda_nitsche  = 0.5 * (mu[0]+mu[1]) * 20 * order * order

    return [ kappa, lambda_nitsche ]

def AssembleCutPoisson(**kwargs):
    ci = kwargs['ci']
    hasneg      = ci.GetElementsOfType(HASNEG)  # <- "hasneg": has (also) negative level set values
    haspos      = ci.GetElementsOfType(HASPOS)  # <- "haspos": has (also) positive level set values    
    hasif       =  BitArray(ci.GetElementsOfType(IF))
    ba_facets   = [GetFacetsWithNeighborTypes(mesh,a=hasneg,b=hasif),GetFacetsWithNeighborTypes(mesh,a=haspos,b=hasif)]


    lsetp1 = kwargs['lsetp1']
    n_lset  = 1.0/grad(lsetp1).Norm() * grad(lsetp1)
    h       = specialcf.mesh_size
    params = kwargs['params']
    [kappa, lambda_nitsche ] = KappaLambda(params,ci)    

    # expressions of test and trial functions (u and v are tuples):
    CutVh = kwargs['CutVh']
    (u,v) = CutVh.TnT()

    gradu = [grad(ui) for ui in u]
    gradv = [grad(vi) for vi in v]

    mu = params["mu"]
    average_flux_u = sum([- kappa[i] * mu[i] * gradu[i] * n_lset for i in [0,1]])
    average_flux_v = sum([- kappa[i] * mu[i] * gradv[i] * n_lset for i in [0,1]])

    # integration domains
    lset_doms = LsetDoms( levelset, lsetp1, subdivlvl )

    #coef_f = [ x*y, x*y]

    # bilinear forms:

    a = BilinearForm(CutVh, symmetric = False)
    f = LinearForm(CutVh)

    #domain integrals:
    for i in [0,1] :
        a += SymbolicBFI( lset_doms[i], form = mu[i] * gradu[i] * gradv[i] )        

        #rhs
        f += SymbolicLFI( lset_doms[i], form = coef_f[i] * v[i] )

    # Nitsche integrals:
    a += SymbolicBFI(lset_doms[2], form = average_flux_u * (v[0]-v[1])
                                        + average_flux_v * (u[0]-u[1])
                                        + lambda_nitsche / h * (u[0]-u[1]) * (v[0]-v[1]) ) 

    for i in [0,1]:
        # Ghost penalty
        a += SymbolicFacetPatchBFI( form =  params["gamma_stab"]  * 
                                    mu[i] / h / h * 
                                    (u[i] - u[i].Other()) * (v[i] - v[i].Other()), 
                                    skeleton=False, definedonelements=ba_facets[i] )

    # # apply mesh adaptation 
    # if( doDeform ):   
    #     mesh.SetDeformation(deformation)

    # setting up matrix and vector
    a.Assemble()
    f.Assemble()

    #updating with non-homogeneous boundary conditions
    gfu = GridFunction( CutVh )
    #boundary lies in positive part
    gfu.components[1].Set( solution[1], BND )
    f.vec.data -= a.mat * gfu.vec

    return (a,f)

def IfaceUnks(CutVh, ci):
    ifdofs = BitArray(CutVh.ndof)
    ifdofs[:] = False
    hasif = BitArray( ci.GetElementsOfType(IF) )

    for el in CutVh.Elements():
        if( hasif[el.nr] ):
            verts = el.vertices
            for v in verts:
                for unk in CutVh.GetDofNrs(v):                    
                    ifdofs[unk] = True

    ifdofs &= CutVh.FreeDofs()
    return ifdofs

class CutFemSmoother:     
    def __init__(self,a,CutVh,ci):
        self.ifdofs = IfaceUnks(CutVh,ci)
        self.a = a
        self.inv = a.mat.Inverse(self.ifdofs, inverse="sparsecholesky")
        self.preJ = a.mat.CreateSmoother( CutVh.FreeDofs() )
    def Smooth( self, u,rhs ):
        self.preJ.Smooth(u, rhs)
        update = u.CreateVector()
        # this needs to be more efficient 
        # with help of ifdofs ...
        update.data = self.a.mat * u - rhs
        u.data -= self.inv * update




if __name__ == '__main__':
    main()   


from ngsolve import *
from ngsolve.meshes import *
from netgen.geom2d import SplineGeometry

from xfem import *
from math import pi
from xfem.lsetcurv import *

from xfem.lset_spacetime import *

ngsglobals.msg_level = 1

def area_of_a_sphere_ST_error(n_steps = 64, i=3, structured_mesh= True):
    if structured_mesh:
        length = 1
        mesh = MakeStructured2DMesh(quads=False,nx=2**(i),ny=2**(i),mapping= lambda x,y : (2*length*x-length,2*length*y-length))
    else:
        square = SplineGeometry()
        square.AddRectangle([-1,-1],[1,1])
        ngmesh = square.GenerateMesh(maxh=(1/2)**(i-1), quad_dominated=False)
        mesh = Mesh (ngmesh)
    
    time_order = 5
    space_order = time_order
    
    lset_adap_st = LevelSetMeshAdaptation_Spacetime(mesh, order_space = space_order, order_time = time_order,
                                                threshold=10.5, discontinuous_qn=False)
    
    coef_told = Parameter(0)
    coef_delta_t = Parameter(0)
    tref = ReferenceTimeVariable()
    t = coef_told + coef_delta_t*tref
    
    r0 = 0.9
    r = sqrt(x**2+y**2+t**2)
    
    ref_vals = { IF: pi/12*(3*sqrt(3) + 2*pi)*r0**2, NEG: 11/24*pi*r0**3}
    tend = r0/2
    
    #ref_vals = { IF: 0.5*pi**2*r0**2, NEG: 2/3*pi*r0**3}
    #tend = 1
    
    # level set
    levelset= r - r0
    
    fes1 = H1(mesh, order=space_order)
    tfe = ScalarTimeFE(time_order)
    st_fes = SpaceTimeFESpace(fes1,tfe)
    
    
    delta_t = tend/n_steps
    coef_delta_t.Set(delta_t)
    told = 0

    lset_p1 = lset_adap_st.lset_p1
    
    sum_vol = 0
    sum_int = 0
    for i in range(n_steps):
        #SpaceTimeInterpolateToP1(levelset,tref,0.,delta_t,lset_p1) # call for the master spacetime_weihack -- 0 and tend are ununsed parameter
        
        deformation = lset_adap_st.CalcDeformation(levelset,tref,told, delta_t)
        
        mesh.SetDeformation(deformation)
        #mesh.SetDeformation(test_deformation)
        val_vol = Integrate({ "levelset" : lset_p1, "domain_type" : NEG}, CoefficientFunction(1.0), mesh, order= 2*space_order, time_order = 2*time_order)
        val_int = Integrate({ "levelset" : lset_p1, "domain_type" : IF}, CoefficientFunction(1.0), mesh, order = 2*space_order, time_order = 2*time_order)
        #print(val_vol, val_int)
        mesh.UnsetDeformation()
        
        sum_vol += val_vol*delta_t
        sum_int += val_int*delta_t
        
        told = told + delta_t
        coef_told.Set(told)

    print("SUM VOL: ", sum_vol)
    print("VOL: ", ref_vals[NEG])
    vol_err = abs(sum_vol - ref_vals[NEG])
    print("\t\tDIFF: ", vol_err)
    
    print("SUM INT: ", sum_int)
    print("AREA: ", ref_vals[IF])
    int_err = abs(sum_int - ref_vals[IF])
    print("\t\tDIFF: ",int_err)
    return (vol_err, int_err)

l2errors_vol = []
l2errors_int = []
for i in [0,1,2,3]:
    (n_steps,i) =  (2**(i+2), i+1)
    (vol_err, int_err) = area_of_a_sphere_ST_error(n_steps, i, True)
    l2errors_vol.append(vol_err)
    l2errors_int.append(int_err)

eocs_vol = [log(l2errors_vol[i-1]/l2errors_vol[i])/log(2) for i in range(1,len(l2errors_vol))]
print("EOCS (VOL): ", eocs_vol)
print("Average: ", sum(eocs_vol)/len(eocs_vol))

eocs_int = [log(l2errors_int[i-1]/l2errors_int[i])/log(2) for i in range(1,len(l2errors_int))]
print("EOCS (INT): ", eocs_int)
print("Average: ", sum(eocs_int)/len(eocs_int))



from ngsolve import *
from ngsolve.meshes import *
from xfem import *

def test_xfes_ndof_2D():
    mesh = MakeStructured2DMesh(quads=False,nx=4,ny=4,mapping=lambda x,y: (2*x-1,2*y-1))
    lsetp1 = GridFunction(H1(mesh,order=1))
    InterpolateToP1((sqrt(x*x+y*y) - 1.0/3.0),lsetp1)
    Vh = H1(mesh, order=1, dirichlet=[1,2,3,4])
    Vhx = XFESpace(Vh, lsetp1)
    assert Vh.ndof == 25
    assert Vhx.ndof == 7

def test_xfes_ndof_3D():
    mesh = MakeStructured3DMesh(hexes=False,nx=4,ny=4,nz=4,mapping=lambda x,y,z: (2*x-1,2*y-1,2*z-1))
    lsetp1 = GridFunction(H1(mesh,order=1))
    InterpolateToP1((sqrt(x*x+y*y) - 1.0/3.0),lsetp1)
    Vh = H1(mesh, order=1, dirichlet=[1,2,3,4])
    Vhx = XFESpace(Vh, lsetp1)
    assert Vh.ndof == 125
    assert Vhx.ndof == 35



import pytest
# the constant pi
from math import pi
# ngsolve stuff
from ngsolve.meshes import *
from ngsolve import *
# basic xfem functionality
from xfem import *
from xfem.lsetcurv import *
# basic geometry features (for the background mesh)

@pytest.mark.parametrize("quad", [True,False])
@pytest.mark.parametrize("order", [1,2,3])

def test_nxfem(quad, order):
  mesh = MakeStructured2DMesh(quads = quad, nx=40, ny=40, mapping = lambda x,y : (3*x-1.5,3*y-1.5))
      
  # manufactured solution and corresponding r.h.s. data CoefficientFunctions:
  r44 = (x*x*x*x+y*y*y*y)
  r41 = sqrt(sqrt(x*x*x*x+y*y*y*y))
  r4m3 = (1.0/(r41*r41*r41))
  r66 = (x*x*x*x*x*x+y*y*y*y*y*y)
  r63 = sqrt(r66)
  r22 = (x*x+y*y)
  r21 = sqrt(r22)
  
  
  solution = [1.0+pi/2.0-sqrt(2.0)*cos(pi/4.0*r44),pi/2.0*r41]
  coef_f = [ (-1.0*sqrt(2.0)*pi*(pi*cos(pi/4*(r44))*(r66)+3*sin(pi/4*(r44))*(r22))),
            (-2.0*pi*3/2*(r4m3)*(-(r66)/(r44)+(r22))) ]
  
  # diffusion cofficients for the subdomains (NEG/POS):
  alpha = [1.0,2.0]
  
  # level set function of the domain (phi = ||x||_4 - 1):
  levelset = (sqrt(sqrt(x*x*x*x+y*y*y*y)) - 1.0)
  
  # class to compute the mesh transformation needed for higher order accuracy
  #  * order: order of the mesh deformation function
  #  * threshold: barrier for maximum deformation (to ensure shape regularity)
  lsetmeshadap = LevelSetMeshAdaptation(mesh, order=order, threshold=0.2, discontinuous_qn=True)
  
  # compute the mesh transformation (applied later)
  deformation = lsetmeshadap.CalcDeformation(levelset)
  
  # the piecewise linear interpolation used in the level set adaptation classe
  lsetp1 = lsetmeshadap.lset_p1
  
  # Gathering information on cut elements:
  #  * domain of (volume/boundary) element:
  #    * NEG= only negative level set values
  #    * POS= only positive level set values
  #    * IF= cut element (negative and positive) level set values
  #  * cut ratio:
  #    If element is cut this describes the ratio between the measure of part in the negative domain
  #    and the measure of the full element.
  ci = CutInfo(mesh, lsetp1)
  
  # extended FESpace 
  
  Vh = H1(mesh, order=order, dirichlet=[1,2,3,4])
  Vhx = XFESpace(Vh, lsetp1)
  VhG = FESpace([Vh,Vhx])
  
  # coefficients / parameters: 
  
  n = 1.0/grad(lsetp1).Norm() * grad(lsetp1)
  h = specialcf.mesh_size
  
  # the cut ratio extracted from the cutinfo-class
  kappa = (CutRatioGF(ci),1.0-CutRatioGF(ci))
  # Nitsche stabilization parameter:
  stab = 10*(alpha[1]+alpha[0])*(order+1)*order/h
  
  # expressions of test and trial functions:
  
  u_std, u_x = VhG.TrialFunction()
  v_std, v_x = VhG.TestFunction()
  
  u = [u_std + op(u_x) for op in [neg,pos]]
  v = [v_std + op(v_x) for op in [neg,pos]]
  
  gradu = [grad(u_std) + op(u_x) for op in [neg_grad,pos_grad]]
  gradv = [grad(v_std) + op(v_x) for op in [neg_grad,pos_grad]]
  
  average_flux_u = sum([- kappa[i] * alpha[i] * gradu[i] * n for i in [0,1]])
  average_flux_v = sum([- kappa[i] * alpha[i] * gradv[i] * n for i in [0,1]])
  
  # Integration domains for integration on negative/positive subdomains and on the interface:
  # Here, the integration is (geometrically) exact if the "levelset"-argument is a piecewise
  # (multi-)linear function. The integration order is chosen according to the arguments in the
  # multilinear forms (but can be overwritten with "force_intorder" in the integration domain). If the
  # "levelset"-argument is not a (multi-)linear function, you can use the "subdivlvl" argument to add
  # additional refinement levels for the geometry approximation. 
  lset_neg = { "levelset" : lsetp1, "domain_type" : NEG, "subdivlvl" : 0}
  lset_pos = { "levelset" : lsetp1, "domain_type" : POS, "subdivlvl" : 0}
  lset_if  = { "levelset" : lsetp1, "domain_type" : IF , "subdivlvl" : 0}
  
  # bilinear forms:
  
  a = BilinearForm(VhG, symmetric = True)
  # l.h.s. domain integrals:
  a += SymbolicBFI(levelset_domain = lset_neg, form = alpha[0] * gradu[0] * gradv[0])
  a += SymbolicBFI(levelset_domain = lset_pos, form = alpha[1] * gradu[1] * gradv[1])
  # Nitsche integrals:
  a += SymbolicBFI(levelset_domain = lset_if , form =       average_flux_u * (v[0]-v[1])
                                                      +     average_flux_v * (u[0]-u[1])
                                                      + stab * (u[0]-u[1]) * (v[0]-v[1]))
  
  f = LinearForm(VhG)
  # r.h.s. domain integrals:
  f += SymbolicLFI(levelset_domain = lset_neg, form = coef_f[0] * v[0])
  f += SymbolicLFI(levelset_domain = lset_pos, form = coef_f[1] * v[1])
  
  # solution vector
  gfu = GridFunction(VhG)
  
  # setting domain boundary conditions:
  gfu.components[0].Set(solution[1], BND)
  
  # activate the mesh deformation (changes the integrals via corresponding transformation):
  mesh.SetDeformation(deformation)
  
  # setting up matrix and vector
  a.Assemble();
  f.Assemble();
  
  # homogenization of boundary data and solution of linear system
  rhs = gfu.vec.CreateVector()
  rhs.data = f.vec - a.mat * gfu.vec
  update = gfu.vec.CreateVector()
  update.data = a.mat.Inverse(VhG.FreeDofs()) * rhs;
  gfu.vec.data += update
  
  # visualization of (discrete) solution: Wherever (interpolated) level set function is negative
  # visualize the first component, where it is positive visualize the second component
  u_coef = gfu.components[0] + IfPos(lsetp1, pos(gfu.components[1]), neg(gfu.components[1]))
  u = [gfu.components[0] + op(gfu.components[1]) for op in [neg,pos]]
  
  err_sqr_coefs = [ (u[i] - solution[i])*(u[i] - solution[i]) for i in [0,1] ]
  
  l2error = sqrt( Integrate( levelset_domain=lset_neg, cf=err_sqr_coefs[0], mesh=mesh,
                             order=2*order, heapsize=1000000)
                  + Integrate(levelset_domain=lset_pos, cf=err_sqr_coefs[1], mesh=mesh,
                              order=2*order, heapsize=1000000))

  reference_error_l2 = { (True,1) : 6e-3,
                         (True,2) : 2e-4,  
                         (True,3) : 6e-6,
                         (False,1) : 8e-3,
                         (False,2) : 2e-4,
                         (False,3) : 7e-6,
  }
  reference_error_dist = { (True,1) : 6e-4,
                           (True,2) : 2e-5,
                           (True,3) : 2e-6,
                           (False,1) : 1e-3,
                           (False,2) : 7e-5,
                           (False,3) : 8e-6,  }
  disterror = lsetmeshadap.CalcMaxDistance(levelset)
  
  print("L2 error : ",l2error)
  print("dist error : ",disterror)
  
  assert l2error < reference_error_l2[(quad,order)]
  assert disterror < reference_error_dist[(quad,order)]
  
  mesh.UnsetDeformation()



# unfitted Heat equation with Neumann b.c.
# solved with a P1-DG-in-time space-time discretization
from ngsolve import *
from netgen.geom2d import unit_square

from xfem import *
from math import pi
import sys

from xfem.lset_spacetime import *

use_sympy = True

if use_sympy:
    from kite_in_sympy import get_Laplace_u_str, get_dt_rho

ngsglobals.msg_level = 1

i = 3

if hasattr(sys, 'argv') and len(sys.argv) == 3 and sys.argv[1] == "i":
    i = int(sys.argv[2])
    print("Sys argv :", sys.argv)
    print("Loading manual val for i: ", i)
else:
    print("Loading default val for i: ", i)

square = SplineGeometry()
square.AddRectangle([-3.5,-1.5],[3.5,1.5])
ngmesh = square.GenerateMesh(maxh=(1/2)**i, quad_dominated=False)
mesh = Mesh (ngmesh)

coef_told = Parameter(0)
coef_delta_t = Parameter(0)
tref = ReferenceTimeVariable()
t = coef_told + coef_delta_t*tref

r0 = 1

rho =  (1 - 3*y**2)*t*sin(pi*t)
rhoL = lambda t: (1- 3*y**2)*t*sin(pi*t)
rho_str = "(1 - 3*y**2)*t*sin(pi*t)"

#convection velocity:
if use_sympy:
    d_rho = eval(get_dt_rho(rho_str))
else:
    rho = (1 - y**2)*t
    rhoL = lambda t: (1 - y**2)*t
    rho_str = "(1 - y**2)*t"
    d_rho = CoefficientFunction (1 - y**2)

w = CoefficientFunction((d_rho,0))

# level set
r = sqrt((x- rho)**2+y**2)
levelset= r - r0
levelsetL = lambda t: sqrt((x- rhoL(t))**2+y**2) - r0

Draw(levelsetL(0.), mesh,"lset", autoscale=False, min = 0, max = 0)

Q = pi/r0   
u_exact = cos(Q*r) * sin(pi*t)
u_exactL = lambda t: cos(Q*sqrt((x- rhoL(t))**2+y**2)) * sin(pi*t)
u_str = "cos(Q*r) * sin(pi*t)"

if use_sympy:
    Laplaceu = eval(get_Laplace_u_str(rho_str, u_str))
else:
    drdx = (x-rho)/r
    dr2dx2 = y**2/r**3

    drdy = y/r*(1 + 2*(x-rho)*t)
    dr2dy2 = (r - y*drdy)/r**2*(1+2*(x-rho)*t) + 4*t**2*y**2/r

    Laplaceu = - Q*sin(pi*t)*(cos(Q*r)*Q* ( drdx**2 + drdy**2) + sin(Q*r)*( dr2dx2 + dr2dy2 ))

coeff_f = -Laplaceu + pi * cos(Q*r) * cos(pi*t)
u_init = u_exact

# polynomial order in time
k_t = 1
# polynomial order in space
k_s = 1
# spatial FESpace for solution
fes1 = H1(mesh, order=k_s)
# polynomial order in time for level set approximation
lset_order_time = 1
# integration order in time
time_order = 2
# time finite element (nodal!)
tfe = ScalarTimeFE(k_t)
# space-time finite element space
st_fes = SpaceTimeFESpace(fes1,tfe, flags = {"dgjumps": True})

#Unfitted heat equation example
tend = 1
delta_t = tend/(2**(i+2))
coef_delta_t.Set(delta_t)
tnew = 0
told = 0

lset_p1 = GridFunction(st_fes)

SpaceTimeInterpolateToP1(levelset,tref,lset_p1)

lset_top = CreateTimeRestrictedGF(lset_p1,1.0)
lset_bottom = CreateTimeRestrictedGF(lset_p1,0.0)

gfu = GridFunction(st_fes)

u_last = CreateTimeRestrictedGF(gfu,0)
u_last.Set(u_exactL(0.))

u,v = st_fes.TnT()

h = specialcf.mesh_size

lset_neg = { "levelset" : lset_p1, "domain_type" : NEG, "subdivlvl" : 0}
lset_neg_bottom = { "levelset" : lset_bottom, "domain_type" : NEG, "subdivlvl" : 0}
lset_neg_top = { "levelset" : lset_top, "domain_type" : NEG, "subdivlvl" : 0}

def SpaceTimeNegBFI(form):
    return SymbolicBFI(levelset_domain = lset_neg, form = form, time_order=time_order)

ci = CutInfo(mesh,time_order=time_order)

#n_F = specialcf.normal(mesh.dim)

hasneg_integrators_a = []
hasneg_integrators_f = []
patch_integrators_a = []
hasneg_integrators_a.append(SpaceTimeNegBFI(form = delta_t*grad(u)*grad(v)))
hasneg_integrators_a.append(SymbolicBFI(levelset_domain = lset_neg_top, form = fix_t(u,1)*fix_t(v,1)))
hasneg_integrators_a.append(SpaceTimeNegBFI(form = -u*dt(v)))
hasneg_integrators_a.append(SpaceTimeNegBFI(form = -delta_t*u*InnerProduct(w,grad(v))))
patch_integrators_a.append(SymbolicFacetPatchBFI(form = delta_t*1.05*h**(-2)*(u-u.Other())*(v-v.Other()),
                                                 skeleton=False, time_order=time_order))
#patch_integrators_a.append(SymbolicFacetPatchBFI( InnerProduct(grad(u) - grad(u.Other()), n_F) * InnerProduct(grad(v) - grad(v.Other()), n_F), skeleton=True, time_order=time_order))
hasneg_integrators_f.append(SymbolicLFI(levelset_domain = lset_neg, form = delta_t*coeff_f*v, time_order=time_order)) 
hasneg_integrators_f.append(SymbolicLFI(levelset_domain = lset_neg_bottom,form = u_last*fix_t(v,0)))


a = BilinearForm(st_fes,check_unused=False,symmetric=False)
for integrator in hasneg_integrators_a + patch_integrators_a:
    a += integrator

f = LinearForm(st_fes)

for integrator in hasneg_integrators_f:
    f += integrator

while tend - told > delta_t/2:
    SpaceTimeInterpolateToP1(levelset,tref,lset_p1)
    RestrictGFInTime(spacetime_gf=lset_p1,reference_time=0.0,space_gf=lset_bottom)
    RestrictGFInTime(spacetime_gf=lset_p1,reference_time=1.0,space_gf=lset_top)

    # update markers in (space-time) mesh
    ci.Update(lset_p1,time_order=time_order)

    # re-compute the facets for stabilization:
    ba_facets = GetFacetsWithNeighborTypes(mesh,a=ci.GetElementsOfType(HASNEG),
                                                b=ci.GetElementsOfType(IF))
    # re-evaluate the "active dofs" in the space time slab
    active_dofs = GetDofsOfElements(st_fes,ci.GetElementsOfType(HASNEG))

    # re-set definedonelements-markers according to new markings:
    for integrator in hasneg_integrators_a + hasneg_integrators_f:
        integrator.SetDefinedOnElements(ci.GetElementsOfType(HASNEG))
    for integrator in patch_integrators_a:
        integrator.SetDefinedOnElements(ba_facets)

    # assemble linear system
    a.Assemble()
    f.Assemble()

    # solve linear system
    inv = a.mat.Inverse(active_dofs,inverse="")
    gfu.vec.data =  inv * f.vec
       
    # evaluate upper trace of solution for
    #  * for error evaluation 
    #  * upwind-coupling to next time slab
    RestrictGFInTime(spacetime_gf=gfu,reference_time=1.0,space_gf=u_last)   

    # update time variable (float and ParameterCL)
    told = told + delta_t
    coef_told.Set(told)
    
    Draw(levelsetL(told), mesh,"lset", autoscale=False, min = 0, max = 0)
    #Draw(IfPos(-levelsetL(told), u_exactL(told) ,float('nan')),mesh,"u", autoscale = False, min = -1, max = 1)
    #Draw(IfPos(-levelsetL(told), u_exactL(told) ,float('nan')),mesh,"u", autoscale = False, min = -1, max = 1)
    #Draw(IfPos(-levelsetL(told), u_exactL(told)-u_last, float('nan')),mesh,"error")
    
    # compute error at end of time slab
    l2error = sqrt(Integrate(lset_neg_top,(u_exactL(told) -u_last)**2,mesh))
    # print time and error
    print("t = {0:10}, l2error = {1:20}".format(told,l2error),end="\n")



import json

#policy_file = "/opt/_internal/pipx/venvs/auditwheel/lib/python3.9/site-packages/auditwheel/policy/manylinux-policy.json"
policy_file = "/opt/_internal/pipx/venvs/auditwheel/lib/python3.10/site-packages/auditwheel/policy/manylinux-policy.json"
data = json.load(open(policy_file))
# todo: read libs from netgen.libs directory and remove hashes from names, then add just libmkl_rt.so.2
additional_libs = [
        "libGLU.so.1",
        "libGLX.so.0",
        "libGLdispatch.so.0",
        "libOpenGL.so.0",
        "libXmu.so.6",
        "libbz2.so.1.0.6",
        "libcsg.so",
        "libcsgvis.so",
        "libfontconfig.so.1.11.1",
        "libfreetype.so.6.14.0",
        "libgeom2d.so",
        "libgeom2dvis.so",
        "libgui.so",
        "libinterface.so",
        "libmesh.so",
        "libmkl_rt.so.2",
        "libngcore.so",
        "libnggui.so",
        "libnglib.so",
        "libocc.so",
        "libpng15.so.15.13.0",
        "libstl.so",
        "libstlvis.so",
        "libtcl8.so",
        "libtk8.so",
        "libuuid.so.1.3.0",
        "libvisual.so",
        "libz.so.1.2.7",
        ]

for entry in data:
    if 'manylinux' in entry['name']:
        entry['lib_whitelist'] += additional_libs
        
json.dump(data, open(policy_file, 'w'))



from ngsolve import *
from xfem import *
from ngsolve.meshes import MakeStructured2DMesh

mesh = MakeStructured2DMesh(quads=False, nx=3, ny=2)

levelset = x - 0.8
gflset = GridFunction(H1(mesh))
InterpolateToP1(levelset, gflset)

ci = CutInfo(mesh, gflset)
roots = ci.GetElementsOfType(NEG)
bads = ci.GetElementsOfType(IF)


#fes = L2(mesh, order=0)
fes = H1(mesh, order=2)

u, v = fes.TnT()
h = specialcf.mesh_size

EA = ElementAggregation(mesh, roots, bads)
E = AggEmbedding(EA, fes)

gfu = GridFunction(fes)
Draw(gfu)
baser = E.CreateRowVector()
for i in range(E.width):
    baser[:] = 0
    baser[i] = 1
    gfu.vec.data = E * baser
    Redraw()
    input(i)
print(" -- done -- ")
